#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Check compatibility of configured relational database schema if exists
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

# Extract the following from the configured DB
# - does configured schema exist? 
# - does _vars table exist? 
# - value of the stored PDK_VER variable

SCHEMA_CORE=${SCHEMA_PREFIX}_core

if  [ $DATABASE = "SQLITE" ] && [ -e $SQLITE_DIR/$SCHEMA_CORE.db ]; then

	SCHEMA_EXISTS="TRUE"
	TABLE_EXISTS=$( echo "ATTACH \"${SQLITE_DB_PATH_STUB}{0}.db\" AS core;
	      SELECT 'TRUE' AS result FROM core.sqlite_master WHERE type='table' AND name='_vars';" \
	      | $PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
	if [ $? != 0 ]; then
		msg_err "ERROR: $PYTHON db_sqlite/runsql.py \"SELECT 'TRUE' AS result FROM core.sqlite_master WHERE type='table' AND name='_vars';\" returned error"
		exit 1
	fi

	if [ "$TABLE_EXISTS" = "TRUE" ]; then
		SCHEMA_VER=$( echo "ATTACH \"${SQLITE_DB_PATH_STUB}{0}.db\" AS core;
			SELECT value FROM core._vars WHERE key = 'PDK_VER';" \
		 	| $PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
		if [ $? != 0 ]; then
			msg_err "ERROR: $PYTHON db_sqlite/runsql.py \"SELECT value FROM $SCHEMA_CORE._vars WHERE key = 'PDK_VER';\" returned error"
			exit 1
		fi
	fi

elif [ $DATABASE = "PGSQL" ]; then

	SCHEMA_EXISTS=$( echo "SELECT 'TRUE' AS result FROM pg_namespace WHERE nspname = '{0}';" | $PYTHON $GXD_PDK_DIR/bin/db_pgsql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
	if [ $? != 0 ]; then
		msg_err "ERROR: $PYTHON db_pgsql/runsql.py \"SELECT 'TRUE' AS result FROM pg_namespace WHERE nspname = '${SCHEMA_CORE}';\" returned error"
		exit 1
	fi

	TABLE_EXISTS=$( echo "SELECT 'TRUE' AS result FROM information_schema.tables WHERE table_schema = '{0}' AND table_name = '_vars';" \
					 | $PYTHON $GXD_PDK_DIR/bin/db_pgsql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
	if [ $? != 0 ]; then
		msg_err "ERROR: $PYTHON db_pgsql/runsql.py \"SELECT 'TRUE' AS result FROM information_schema.tables WHERE table_schema = '$SCHEMA_CORE' AND table_name = '_vars';\" returned error"
		exit 1
	fi

	if [ "$TABLE_EXISTS" = "TRUE" ]; then
		SCHEMA_VER=$( echo "SELECT \"value\" FROM {0}._vars WHERE \"key\" = 'PDK_VER';" \
						 | $PYTHON $GXD_PDK_DIR/bin/db_pgsql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
		if [ $? != 0 ]; then
			msg_err "ERROR: $PYTHON db_pgsql/runsql.py \"SELECT \"value\" FROM $SCHEMA_CORE._vars WHERE \"key\" = 'PDK_VER';\" returned error"
			exit 1
		fi
	fi

elif [ $DATABASE = "MYSQL" ]; then

	SCHEMA_EXISTS=$( echo "SELECT 'TRUE' AS result FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '{0}';" | $PYTHON $GXD_PDK_DIR/bin/db_mysql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
	if [ $? != 0 ]; then
		msg_err "ERROR: $PYTHON db_mysql/runsql.py \"SELECT 'TRUE' AS result FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '${SCHEMA_CORE}';\" returned error"
		exit 1
	fi	

	TABLE_EXISTS=$( echo "SELECT 'TRUE' AS result FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = '{0}' AND TABLE_NAME = '_vars';" \
					 | $PYTHON $GXD_PDK_DIR/bin/db_mysql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
	if [ $? != 0 ]; then
		msg_err "ERROR: $PYTHON db_mysql/runsql.py \"SELECT 'TRUE' AS result FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = '$SCHEMA_CORE' AND TABLE_NAME = '_vars';\" returned error"
		exit 1
	fi

	if [ "$TABLE_EXISTS" = "TRUE" ]; then
		SCHEMA_VER=$( echo "SELECT \`value\` FROM {0}._vars WHERE \`key\` = 'PDK_VER';" \
						 | $PYTHON $GXD_PDK_DIR/bin/db_mysql/runsql.py "" "$SCHEMA_CORE" FALSE ";" "FETCH_ONE" )
		if [ $? != 0 ]; then
			msg_err "ERROR: $PYTHON db_mysql/runsql.py \"SELECT \`value\` FROM $SCHEMA_CORE._vars WHERE \`key\` = 'PDK_VER';\" returned error"
			exit 1
		fi
	fi
fi

# Check schema compatibility

if  [ "$SCHEMA_EXISTS" = "TRUE" ]; then

	if  [ "$TABLE_EXISTS" = "TRUE" ] && [ "$SCHEMA_VER" != "" ]; then

		if [ "$SCHEMA_VER" != "$EARLIEST_COMPATIBLE_RDB_SCHEMA" ]; then

			# Figure out lesser version between current DB schema version and PDK earliest compatible schema
			LESSER_VER=$( echo "$SCHEMA_VER|$EARLIEST_COMPATIBLE_RDB_SCHEMA" | tr '|' '\n' | sort -V | tr '\n' '|' | cut -d'|' -f1 )

			# Error out if current DB schema version the lesser one
			if [ "$LESSER_VER" = "$SCHEMA_VER" ]; then
				msg_err "ERROR: Earliest compatible schema version ($EARLIEST_COMPATIBLE_RDB_SCHEMA) exceeds current database schema version ($SCHEMA_VER)"
				echo
				exit 1
			fi
		fi

	else 	# _vars table does not exist or PDK_VER variable not populated, potentially older schema, error out

		msg_err "ERROR: Earliest compatible schema version is $EARLIEST_COMPATIBLE_RDB_SCHEMA, but current database schema version is not specified"
		echo "If you're certain DB schemas are compatible, use -s option to skip DB schema version check"
		echo
		exit 1

	fi

fi
