import os
import sys
#from timeit import default_timer as timer
from pymongo import MongoClient, DeleteMany, DeleteOne
from pymongo.errors import PyMongoError

sys.path.append(os.path.join(sys.path[0],'..'))
import gxd_functions as gxd

OPS_PER_WRITE = 1000

schema_prefix	= sys.argv[1]
collection		= sys.argv[2]
core_action 	= sys.argv[3]

if len(sys.argv) > 4 and sys.argv[4].upper() == 'TRUE':
	do_drop_core = True
else:
	do_drop_core = False

gxd_dir 		= os.environ['GXD_PDK_DIR']
custom_conf 	= os.environ['GXD_PDK_CONFIG']
ingest_id 		= os.environ['GXD_INGEST_JOB_ID']

mongo_host 	= gxd.get_config_setting('MONGO_HOST', gxd_dir, custom_conf)
mongo_port 	= gxd.get_config_setting('MONGO_PORT', gxd_dir, custom_conf)
mongo_user 	= gxd.get_config_setting('MONGO_USER', gxd_dir, custom_conf)
mongo_pass 	= gxd.get_config_setting('MONGO_PASS', gxd_dir, custom_conf)

fields_tsv 	= os.path.join( gxd_dir, "config", gxd.get_config_setting('FIELD_CONFIGS', gxd_dir, custom_conf) )
fields 		= gxd.get_fields(fields_tsv, collection)
doc_key 	= gxd.get_pdk_field_name(fields, 'PRIMARY_KEY')
ingest_id 	= gxd.get_pdk_field_name(fields, 'INGEST_ID')

db_deletes	= gxd.get_config_setting('DB_DELETES', gxd_dir, custom_conf)
bool_to_int	= gxd.get_config_setting('BOOL_TO_INT', gxd_dir, custom_conf)
delete_flag = gxd.get_pdk_field_name(fields, 'DELETE_FLAG')

schema_stage	= f'{schema_prefix}_stage'
schema_core		= f'{schema_prefix}_core'

try:
	db_stage = MongoClient( "{}:{}".format(mongo_host, mongo_port), username=mongo_user, password=mongo_pass )[schema_stage]
	db_core  = MongoClient( "{}:{}".format(mongo_host, mongo_port), username=mongo_user, password=mongo_pass )[schema_core]
except pymongo.errors.PyMongoError as e:
	gxd.error("\t{0}: exception connecting to Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
	sys.exit(1)

coll_stage	= db_stage.get_collection(collection)
coll_core	= db_core.get_collection(collection)

# Skip if stage collection is empty (analogous to RDB flow)
doc_cnt = coll_stage.count_documents({})
if doc_cnt == 0:
	gxd.ok("\tCollection '{}' is empty in stage, dropping namesake core collection..\n".format(collection))
	coll_core.drop()
	sys.exit()
else:
	gxd.ok("\tCollection '{}' has {} document(s) in stage\n".format(collection, doc_cnt))

if core_action.upper() == 'REPLACE' and do_drop_core:
	gxd.ok("\tDropping core collection..\n")
	coll_core.drop()

# Delete stage records with delete flag prior to replacing core
if core_action.upper() == 'REPLACE' and db_deletes and delete_flag:
	delete_flag_value = '1' if bool_to_int else 'TRUE'
	try:
		gxd.ok("\tRemoving stage documents marked as deleted..\n")
		coll_stage.delete_many({ delete_flag : delete_flag_value })
	except pymongo.errors.PyMongoError as e:
		gxd.error("\t{0}: exception deleting from collection in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)


# Use aggregation pipeline to update/insert records into namesake collection in core schema
pipeline = []
pipeline.append({ '$set': 	{ '_id': f'${ doc_key }' } })
pipeline.append({ '$merge': {
						'into': { 'db': f'{ schema_core }', 'coll': f'{ collection }' },
						'on': '_id',
						'whenMatched': 'replace',
						'whenNotMatched': 'insert' }
				})
try:
	gxd.ok("\tInserting/updating documents in core via aggregation pipeline..\n")
	coll_stage.aggregate(pipeline)
except pymongo.errors.PyMongoError as e:
	gxd.error("\t{0}: exception aggregating pipeline in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
	sys.exit(1)


# Delete core records with delete flag after updating core
if core_action.upper() == 'UPDATE' and db_deletes and delete_flag:
	delete_flag_value = '1' if bool_to_int else 'TRUE'
	try:
		gxd.ok("\tRemoving core documents marked as deleted..\n")
		coll_core.delete_many({ delete_flag : delete_flag_value })
	except pymongo.errors.PyMongoError as e:
		gxd.error("\t{0}: exception deleting from collection in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)


# Are we replacing over existing core (vs dropping it first)? - if so, we also need to delete core documents not having current ingest_id
if core_action.upper() == 'REPLACE' and not do_drop_core:

	gxd.ok("\tProceeding to remove old documents in core\n")

	# Use aggregation pipeline to determine current ingest identifier in stage collection (must be unique)
	pipeline = []
	pipeline.append({ '$group': { '_id': f'${ ingest_id }' } })

	try:
		gxd.ok("\tFinding ingest identifier in stage via aggregation pipeline..\n")
		ingest_id_values = [ i['_id'] for i in coll_stage.aggregate(pipeline, allowDiskUse=True) ]
	except pymongo.errors.PyMongoError as e:
		gxd.error("\t{0}: exception aggregating pipeline in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)


	# Check if stage collection contains more than one ingest identifier
	if len(ingest_id_values) > 1:
		gxd.error("\t{0}: stage collection '{1}' contains more than one ingest identifier: {2}\n".format( os.path.basename(__file__), collection, ingest_id_values ))
		sys.exit(1)


	# Use aggregation pipeline to find core documents not having current ingest identifier
	pipeline = []
	pipeline.append({ '$match': { f'{ ingest_id }': { '$ne': f'{ ingest_id_values[0] }' } } })
	pipeline.append({ '$project': { '_id': 1 } })

	try:
		gxd.ok("\tFinding any core documents not having current stage ingest identifier ({}) in core via aggregation pipeline..\n".format(ingest_id_values[0]))
		deletes = list(coll_core.aggregate(pipeline))
	except pymongo.errors.PyMongoError as e:
		gxd.error("\t{0}: exception aggregating pipeline in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)

	if len(deletes) > 0:

		# Build bulk write to delete the documents not having current ingest identifier
		gxd.ok("\tBuilding bulk delete operation..\n")
		operations = []
		for doc in deletes:
			operations.append( DeleteOne({ '_id': doc['_id'] }) )


		# Execute the bulk delete
		write_cnt = 0
		gxd.ok("\tExecuting bulk delete operation..\n")
		for write in gxd.group_iterator(operations, OPS_PER_WRITE):
			write_cnt += 1
			try:
				coll_core.bulk_write( write, ordered=False )
				sys.stdout.write("\r\tDatabase writes ({0} documents/operations each): {1} ".format( OPS_PER_WRITE, write_cnt))
			except pymongo.errors.PyMongoError as e:
				gxd.error("\t{0}: exception writing to Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
				sys.exit(1)

		sys.stdout.write("\n")
		gxd.log("\tDatabase writes ({0} documents/operations each): {1}\n".format( OPS_PER_WRITE, write_cnt))


# Drop stage collection when we're done
gxd.ok("\tDropping stage collection..\n")
coll_stage.drop()


#start = timer()
# # Code to time
#end = timer()
#print(int((end-start)*1000))

