import os
import sys
from pymongo import MongoClient, DeleteMany
from pymongo.errors import PyMongoError

sys.path.append(os.path.join(sys.path[0],'..'))
import gxd_functions as gxd

OPS_PER_WRITE = 1000

mongo_schema 	= sys.argv[1]
collection		= sys.argv[2]

gxd_dir 		= os.environ['GXD_PDK_DIR']
custom_conf 	= os.environ['GXD_PDK_CONFIG']
ingest_id 		= os.environ['GXD_INGEST_JOB_ID']

mongo_host 	= gxd.get_config_setting('MONGO_HOST', gxd_dir, custom_conf)
mongo_port 	= gxd.get_config_setting('MONGO_PORT', gxd_dir, custom_conf)
mongo_user 	= gxd.get_config_setting('MONGO_USER', gxd_dir, custom_conf)
mongo_pass 	= gxd.get_config_setting('MONGO_PASS', gxd_dir, custom_conf)

fields_tsv 	= os.path.join( gxd_dir, "config", gxd.get_config_setting('FIELD_CONFIGS', gxd_dir, custom_conf) )
fields 		= gxd.get_fields(fields_tsv, collection)
doc_key 	= gxd.get_pdk_field_name(fields, 'PRIMARY_KEY')
update_id 	= gxd.get_pdk_field_name(fields, 'UPDATE_ID')

try:
	db = MongoClient( "{}:{}".format(mongo_host, mongo_port), username=mongo_user, password=mongo_pass )[mongo_schema]
except pymongo.errors.PyMongoError as e:
	gxd.error("\t{0}: exception connecting to Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
	sys.exit(1)

coll = db.get_collection(collection)

# Skip if stage collection is empty
if coll.count_documents({}) == 0:
	gxd.ok("\tCollection {} is empty\n".format(collection))
	sys.exit()

# Use aggregation pipeline to sort by discriminator field and identify duplicate doc_keys
pipeline = []
pipeline.append({ '$sort': { update_id: 1 } })
pipeline.append({ '$group': { 	'_id': f'${ doc_key }',
								'ids': { '$push': '$_id' },
								'count': { '$sum': 1 } } 
				})
pipeline.append({ '$match': { 'count': { '$gt': 1 } } })
#pipeline.append({ '$facet': { 'meta': [{ '$count':'count' }], 'data': [{ '$skip':0 }] }})

try:
	gxd.ok("\tIdentifying duplicates via aggregation pipeline..\n")
	duplicates = coll.aggregate(pipeline, allowDiskUse=True)
except pymongo.errors.PyMongoError as e:
	gxd.error("\t{0}: exception aggregating pipeline in Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
	sys.exit(1)

# Iterate over Mongo cursor and populate elements to a list
duplicates = [ doc for doc in duplicates ]
dupe_count = len(duplicates)

if dupe_count > 0:

	# Remove objects with max update_id (last) from resulting set and build bulk write to delete the rest
	# This should also have an effect of removing any 'pure' duplicates (analogous to MD5 deduping in RDB flow)
	gxd.ok("\tBuilding bulk delete operation for {} duplicates..\n".format(dupe_count))
	operations = []
	for dupe in duplicates:
		dupe['ids'].pop()		# Remove the last object (having max update_id) from the list of delete candidates
		operations.append( DeleteMany({ '_id': { '$in': dupe['ids'] } }) )

	# Execute the bulk delete
	gxd.ok("\tExecuting bulk delete operation for duplicates..\n")
	write_count = 0
	for write in gxd.group_iterator(operations, OPS_PER_WRITE):
		write_count += 1
		try:
			coll.bulk_write( write, ordered=False )
			sys.stdout.write("\r\tDatabase writes ({} documents/operations each): {} ".format( OPS_PER_WRITE, write_count))
		except pymongo.errors.PyMongoError as e:
			gxd.error("\t{0}: exception writing to Mongo database: {1}\n".format( os.path.basename(__file__), str(e) ))
			sys.exit(1)

	sys.stdout.write("\n")
	gxd.log("\tDatabase writes ({0} documents each): {1}\n".format( OPS_PER_WRITE, write_count))

else:
	gxd.ok("\tNo duplicates found\n")