#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Reads rows from TSV file for a given table, injecting ingest_id into each row
#  Runs parametrized inserts against configured database using transactions
#  Each transactions contains certain number of statements defined below
#  Each statement contains certain number of rows defined below
#  =============================================================================

import os
import sys
import mysql.connector

sys.path.append(os.path.join(sys.path[0],'..'))
import gxd_functions as gxd

ROWS_PER_INSERT 			= 100
STATEMENTS_PER_TRANSACTION 	= 100

table 			= sys.argv[1]
tsv_path 		= sys.argv[2]
mysql_schema 	= sys.argv[3]

gxd_dir 		= os.environ['GXD_PDK_DIR']
custom_conf 	= os.environ['GXD_PDK_CONFIG']
ingest_id 		= os.environ['GXD_INGEST_JOB_ID']

fields_tsv 		= os.path.join( gxd_dir, "config", gxd.get_config_setting('FIELD_CONFIGS', gxd_dir, custom_conf) )
null_value 		= gxd.get_config_setting('NULL_VALUE', gxd_dir, custom_conf)

mysql_host 		= gxd.get_config_setting('MYSQL_HOST', gxd_dir, custom_conf)
mysql_port 		= gxd.get_config_setting('MYSQL_PORT', gxd_dir, custom_conf)
mysql_user 		= gxd.get_config_setting('MYSQL_USER', gxd_dir, custom_conf)
mysql_pass 		= gxd.get_config_setting('MYSQL_PASS', gxd_dir, custom_conf)

fields = gxd.get_fields(fields_tsv, table)
field_names = [ None ] * len(fields)
for name, field in fields.items():
	field_names[field.get('idx')] = name

field_names_str  = ",".join([ '`{}`'.format(f) for f in field_names ])
field_params_str = ",".join([ '%s' for f in fields ])



injections = []

# Use Python function attribute to create auto-counter that increments with every function call
def counter():
	if not hasattr(counter,'value'):
		counter.value = 0
	counter.value += 1
	return counter.value

row_ordinal_field_name = gxd.get_pdk_field_name(fields, 'BATCH_UPDATE_ORD')
if row_ordinal_field_name:
	idx_row_ordinal = fields.get(row_ordinal_field_name).get('idx')
	injections.append( { 'index': idx_row_ordinal, 'function': counter } )

ingest_id_field_name = gxd.get_pdk_field_name(fields, 'INGEST_ID')
if ingest_id_field_name:
	idx_ingest_id = fields.get(ingest_id_field_name).get('idx')
	injections.append( { 'index': idx_ingest_id, 'function': lambda : ingest_id } )



try:
	db = mysql.connector.connect(host = mysql_host, port = mysql_port, user = mysql_user, password = mysql_pass, database = mysql_schema)
except mysql.connector.Error as e:
	gxd.error("\t{0}: DBAPI exception connecting to MySQL database: {1}\n".format( os.path.basename(__file__), str(e) ))
	sys.exit(1)

tsv = gxd.tsv_iterator( os.path.join(tsv_path, table + ".tsv"), null_value = null_value, injections = injections )

#print "Inserting into {}".format(table)
sql_statement = "INSERT INTO {0} ({1}) VALUES ({2})".format(table, field_names_str, field_params_str)

statement_counter = 0
transaction_counter = 0
for group in gxd.group_iterator(tsv, ROWS_PER_INSERT):
	try:
		db.cursor().executemany( sql_statement, [ tuple(row) for row in group ] )
		statement_counter += 1
	except mysql.connector.Error as e:
		gxd.error("\t{0}: DBAPI exception executing insert statements: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)

	if statement_counter == STATEMENTS_PER_TRANSACTION:
		statement_counter = 0
		transaction_counter += 1
		try:
			db.commit()
			sys.stdout.write("\r\tDatabase transactions ({} rows each): {} ".format( ROWS_PER_INSERT * STATEMENTS_PER_TRANSACTION, transaction_counter))
		except mysql.connector.Error as e:
			gxd.error("\t{0}: DBAPI exception running database transaction: {1}\n".format( os.path.basename(__file__), str(e) ))
			sys.exit(1)

if statement_counter > 0:
	transaction_counter += 1
	try:
		db.commit()
		sys.stdout.write("\r\tDatabase transactions ({} rows each): {} ".format( ROWS_PER_INSERT * STATEMENTS_PER_TRANSACTION, transaction_counter))
	except mysql.connector.Error as e:
		gxd.error("\t{0}: DBAPI exception running database transaction: {1}\n".format( os.path.basename(__file__), str(e) ))
		sys.exit(1)

sys.stdout.write("\n")
gxd.log("\tDatabase transactions ({0} rows each): {1}\n".format( ROWS_PER_INSERT * STATEMENTS_PER_TRANSACTION, transaction_counter))
