#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Core replace sequence for SQLite
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

VARS=$GXD_PDK_DIR/vars
GXD_TSV_DIR=$GXD_PDK_DIR/files/tsv
echo "JOB_PHASE=CORE_REPLACE" > $VARS/job_phase.var

SCHEMA_CORE=${SCHEMA_PREFIX}_core
SCHEMA_STAGE=${SCHEMA_PREFIX}_stage

echo

# Check which step the script failed on during previous run, if any
# If it is not 0, we'll skip to the failed step
JOB_STATUS=$1
if  [ $JOB_STATUS = "ERROR" ] &&
	[ -e $VARS/job_phase_step.var ]
then
	. $VARS/job_phase_step.var
else
	JOB_PHASE_STEP=0
fi

while read LINE; do
	PDK_VER=$LINE
done < $GXD_PDK_DIR/docs/version.txt

TRUTH_VALUE="TRUE"

# 1. Create schema
if [ $JOB_PHASE_STEP -le 1 ]; then
	echo "JOB_PHASE_STEP=1" > $VARS/job_phase_step.var

	if [ -e $VARS/last_good_ingest_id.var ]; then
		. $VARS/last_good_ingest_id.var
	fi

	# Check if schema exists first
	if [ -e $SQLITE_DIR/$SCHEMA_CORE.db ]; then
		SCHEMA_EXISTS="TRUE"
	fi

	if [ "$SCHEMA_EXISTS" != "TRUE" ]; then
		msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/schema_core.sql $SCHEMA_CORE"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/schema_core.sql $SCHEMA_CORE
		if [ $? != 0 ]; then
			msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/schema_core.sql $SCHEMA_CORE returned error"
			exit_abnormal
		fi

		# Create _vars table and initialize variables (PDK version)
		msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/vars.sql '$SCHEMA_CORE|$PDK_VER'"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/vars.sql "$SCHEMA_CORE|$PDK_VER"
		if [ $? != 0 ]; then
			msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/vars.sql '$SCHEMA_CORE|$PDK_VER' returned error"
			exit_abnormal
		fi
	fi
fi

# 2. Disable FKs and drop core indexes
if [ $JOB_PHASE_STEP -le 2 ]; then
	echo "JOB_PHASE_STEP=2" > $VARS/job_phase_step.var
	msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/disable_fks_drop_indexes.sql $SCHEMA_CORE"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/replace_core/disable_fks_drop_indexes.sql $SCHEMA_CORE
	if [ $? != 0 ]; then
		msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/disable_fks_drop_indexes.sql $SCHEMA_CORE returned error"
		exit_abnormal
	fi
fi

# 3. Drop stage indexes (we'll be truncating stage data after successful copy to core)
if [ $JOB_PHASE_STEP -le 3 ]; then
	echo "JOB_PHASE_STEP=3" > $VARS/job_phase_step.var
	msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE
	if [ $? != 0 ]; then
		msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE returned error"
		exit_abnormal
	fi
fi

# 4. Delete stage objects flagged for deletion
if [ $JOB_PHASE_STEP -le 4 ]; then
	echo "JOB_PHASE_STEP=4" > $VARS/job_phase_step.var
	if [ $DB_DELETES = "TRUE" ]; then
		msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/stage/remove_deleted.sql '$SCHEMA_STAGE|$TRUTH_VALUE'"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/stage/remove_deleted.sql "$SCHEMA_STAGE|$TRUTH_VALUE"
		if [ $? != 0 ]; then
			msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/stage/remove_deleted.sql '$SCHEMA_STAGE|$TRUTH_VALUE' returned error"
			exit_abnormal
		fi
	fi
fi

# 5. Move stage data to core if exists in stage
if [ $JOB_PHASE_STEP -le 5 ]; then
	echo "JOB_PHASE_STEP=5" > $VARS/job_phase_step.var

	# Since we're advancing table by table, ensure that tracker is created (or exists) to record successfully processed tables
	touch $VARS/job_phase_step_tracker.var

	TABLES=$( cut -f2 $GXD_PDK_DIR/config/$FIELD_CONFIGS | sort | uniq )
	for TABLE in $TABLES; do

		# Only proceed if the table has not already been successfully processed
		FOUND=$( grep -w $TABLE $VARS/job_phase_step_tracker.var )
		if [ "$FOUND" = "" ]; then

			# Retrieve row count for the current table
			msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py \"SELECT COUNT(*) FROM <$SCHEMA_STAGE>.$TABLE;\""
			ROW_COUNT=$( echo "ATTACH \"${SQLITE_DB_PATH_STUB}{0}.db\" AS stage;
			      SELECT COUNT(*) FROM stage.{1};" | $PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py "" "$SCHEMA_STAGE|$TABLE" FALSE ";" "FETCH_ONE" )
			if [ $? != 0 ]; then
				msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py \"SELECT COUNT(*) FROM <$SCHEMA_STAGE>.$TABLE;\" returned error"
				exit_abnormal
			fi
			msg_ok "REPLACE: -> $ROW_COUNT"

			# If no rows, simply truncate the core table
			if [ "$ROW_COUNT" = "0" ]; then
				msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py \"DELETE FROM <$SCHEMA_CORE>.$TABLE;\""
				ROW_COUNT=$( echo "ATTACH \"${SQLITE_DB_PATH_STUB}{0}.db\" AS core;
					DELETE FROM core.{1};" | $PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py "" "$SCHEMA_CORE|$TABLE" )
				if [ $? != 0 ]; then
					msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py \"DELETE FROM $SCHEMA_CORE.$TABLE;\" returned error"
					exit_abnormal
				fi
			else
				# Otherwise truncate core table, copy data from stage, truncate stage table, in one transaction
				msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/move_stage_to_core.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TABLE'"
				$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/replace_core/move_stage_to_core.sql "$SCHEMA_STAGE|$SCHEMA_CORE|$TABLE" FALSE
				if [ $? != 0 ]; then
					msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/move_stage_to_core.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TABLE' returned error"
					exit_abnormal
				fi
			fi

			# Record table as successfully processed
			echo "$TABLE" >> $VARS/job_phase_step_tracker.var

		else
			msg_ok "REPLACE: Table $TABLE has already been processed"
		fi

	done

	# Remove tracker if all tables have been successfully processed
	rm -f $VARS/job_phase_step_tracker.var
fi

# 6. Recreate core indexes and enable FKs
if [ $JOB_PHASE_STEP -le 6 ]; then
	echo "JOB_PHASE_STEP=6" > $VARS/job_phase_step.var
	msg_ok "REPLACE: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/create_indexes_enable_fks.sql $SCHEMA_CORE"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/replace_core/create_indexes_enable_fks.sql $SCHEMA_CORE TRUE
	if [ $? != 0 ]; then
		msg_err "REPLACE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/replace_core/create_indexes_enable_fks.sql $SCHEMA_CORE returned error"
		exit_abnormal
	fi
fi


# Once all steps ran successfully, clear previously failed step
rm -f $VARS/job_phase_step.var

echo
