#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Core update sequence for SQLite
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

VARS=$GXD_PDK_DIR/vars
GXD_TSV_DIR=$GXD_PDK_DIR/files/tsv
echo "JOB_PHASE=CORE_UPDATE" > $VARS/job_phase.var

SCHEMA_CORE=${SCHEMA_PREFIX}_core
SCHEMA_STAGE=${SCHEMA_PREFIX}_stage

echo

# Check which step the script failed on during previous run, if any
# If it is not 0, we'll skip to the failed step
JOB_STATUS=$1
if  [ $JOB_STATUS = "ERROR" ] &&
	[ -e $VARS/job_phase_step.var ]
then
	. $VARS/job_phase_step.var
else
	JOB_PHASE_STEP=0
fi

while read LINE; do
	PDK_VER=$LINE
done < $GXD_PDK_DIR/docs/version.txt

TRUTH_VALUE="TRUE"

# 1. Create schema
if [ $JOB_PHASE_STEP -le 1 ]; then
	echo "JOB_PHASE_STEP=1" > $VARS/job_phase_step.var

	if [ -e $VARS/last_good_ingest_id.var ]; then
		. $VARS/last_good_ingest_id.var
	fi

	# Check if schema exists first
	if [ -e $SQLITE_DIR/$SCHEMA_CORE.db ]; then
		SCHEMA_EXISTS="TRUE"
	fi

	if [ "$SCHEMA_EXISTS" != "TRUE" ]; then
		msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/schema_core.sql $SCHEMA_CORE"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/schema_core.sql $SCHEMA_CORE
		if [ $? != 0 ]; then
			msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/schema_core.sql $SCHEMA_CORE returned error"
			exit_abnormal
		fi

		# Create _vars table and initialize variables (PDK version)
		msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/vars.sql '$SCHEMA_CORE|$PDK_VER'"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/vars.sql "$SCHEMA_CORE|$PDK_VER"
		if [ $? != 0 ]; then
			msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/vars.sql '$SCHEMA_CORE|$PDK_VER' returned error"
			exit_abnormal
		fi
	fi
fi

# 2. Insert new entities
if [ $JOB_PHASE_STEP -le 2 ]; then
	echo "JOB_PHASE_STEP=2" > $VARS/job_phase_step.var
	# Insert new entities into core (regardless of stage delete or inactive flag value)
	msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/insert_new_all.sql '$SCHEMA_STAGE|$SCHEMA_CORE'"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/update_core/insert_new_all.sql "$SCHEMA_STAGE|$SCHEMA_CORE"
	if [ $? != 0 ]; then
		msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/insert_new_all.sql '$SCHEMA_STAGE|$SCHEMA_CORE' returned error"
		exit_abnormal
	fi
fi

# 3. Update new+existing entities in batches
if [ $JOB_PHASE_STEP -le 3 ]; then
	echo "JOB_PHASE_STEP=3" > $VARS/job_phase_step.var

	# Since we're advancing entity by entity, ensure that tracker is created (or exists) to record successfully processed entities
	touch $VARS/job_phase_step_tracker.var

	while read -r BATCH_CONFIG; do
		ENTITY=$( echo "$BATCH_CONFIG" | cut -f1 )
		BATCH_SIZE=$( echo "$BATCH_CONFIG" | cut -f2 )

		# Only proceed if the entity has not already been successfully processed
		FOUND=$( grep -w $ENTITY $VARS/job_phase_step_tracker.var )
		if [ "$FOUND" = "" ]; then

			# Retrieve max batch ordinal for the current table (need deterministic order for batched updates)
			msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/get_max_batch_ordinal.sql '$SCHEMA_STAGE|$ENTITY'"
			ROW_COUNT=$( $PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py db/sqlite/update_core/get_max_batch_ordinal.sql "$SCHEMA_STAGE|$ENTITY" FALSE ";" "FETCH_ONE" )
			if [ $? != 0 ]; then
				msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/get_max_batch_ordinal.sql '$SCHEMA_STAGE|$ENTITY' returned error"
				exit_abnormal
			fi
			msg_ok "UPDATE: -> $ROW_COUNT"

			# Skip table if no rows
			if [ "$ROW_COUNT" = "0" ]; then
				continue
			else
				# Otherwise calculate required number of update batches and proceed to update
				NUM_BATCHES=$(( (ROW_COUNT - 1) / BATCH_SIZE + 1 ))

				BATCH=0
				# Load any previously stored batch number (in case of previous error) for the entity
				if [ -e $VARS/update_core_${ENTITY}_batch.var ]; then
				    . $VARS/update_core_${ENTITY}_batch.var
				    msg_ok "UPDATE: Found previously stored update batch number for ${ENTITY}, resuming update from batch ${BATCH}"
				fi

				while [ $BATCH -lt $NUM_BATCHES ]; do
					LOWER_BOUND=$((BATCH_SIZE * BATCH))
					UPPER_BOUND=$((LOWER_BOUND + BATCH_SIZE))

					if [ "$UPDATE_FLAGS" = "TRUE" ]; then
						# Update non-deleted, non-inactive entities from stage to core; for deleted, then inactive entities, update corresponding core flag and updateId, updateDate, ingestId fields only
						msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/update_batch_$ENTITY.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND|$TRUTH_VALUE'"
						$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/update_core/update_batch_$ENTITY.sql "$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND|$TRUTH_VALUE" FALSE ";"
						if [ $? != 0 ]; then
							msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/update_batch_$ENTITY.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND|$TRUTH_VALUE' returned error"
							exit_abnormal
						fi
					else
						# Update entities from stage to core in their entirety regardless of flags
						msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/update_batch_all_$ENTITY.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND'"
						$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/update_core/update_batch_all_$ENTITY.sql "$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND" FALSE ";"
						if [ $? != 0 ]; then
							msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/update_batch_all_$ENTITY.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$LOWER_BOUND|$UPPER_BOUND' returned error"
							exit_abnormal
						fi
					fi

					BATCH=$((BATCH + 1))
					echo "BATCH=$BATCH" > $VARS/update_core_${ENTITY}_batch.var
				done
				
				# Remove batch counter if all batches have been successfully processed
				rm -f $VARS/update_core_${ENTITY}_batch.var
			fi

			# Record entity as successfully processed
			echo "$ENTITY" >> $VARS/job_phase_step_tracker.var

		else
			msg_ok "UPDATE: Entity $ENTITY has already been processed"
		fi

	done < $GXD_PDK_DIR/config/$ENTITY_CONFIGS

	# Remove tracker if all entities have been successfully processed
	rm -f $VARS/job_phase_step_tracker.var		
fi

# 4. If processing deletes, delete children in core, for each entity
if [ $JOB_PHASE_STEP -le 4 ]; then
	echo "JOB_PHASE_STEP=4" > $VARS/job_phase_step.var
	if [ $DB_DELETES = "TRUE" ]; then

		while read -r ENTITY_CONFIG; do
			ENTITY=$( echo "$ENTITY_CONFIG" | cut -f1 )

			msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/delete_${ENTITY}_children.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE'"
			$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/update_core/delete_${ENTITY}_children.sql "$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE" FALSE ";"
			if [ $? != 0 ]; then
				msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/delete_${ENTITY}_children.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE' returned error"
				exit_abnormal
			fi

		done < $GXD_PDK_DIR/config/$ENTITY_CONFIGS

	fi
fi

# 5. If processing deletes, delete core entities (updating self-references - if any - to NULL)
if [ $JOB_PHASE_STEP -le 5 ]; then
	echo "JOB_PHASE_STEP=5" > $VARS/job_phase_step.var
	if [ $DB_DELETES = "TRUE" ]; then
		msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/delete_entities.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE'"
		$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/update_core/delete_entities.sql "$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE"
		if [ $? != 0 ]; then
			msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/update_core/delete_entities.sql '$SCHEMA_STAGE|$SCHEMA_CORE|$TRUTH_VALUE' returned error"
			exit_abnormal
		fi
	fi
fi

# 6. Drop stage indexes
if [ $JOB_PHASE_STEP -le 6 ]; then
	echo "JOB_PHASE_STEP=6" > $VARS/job_phase_step.var
	msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE
	if [ $? != 0 ]; then
		msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/stage/drop_indexes.sql $SCHEMA_STAGE returned error"
		exit_abnormal
	fi
fi

# 7. Clear stage
if [ $JOB_PHASE_STEP -le 7 ]; then
	echo "JOB_PHASE_STEP=7" > $VARS/job_phase_step.var
	msg_ok "UPDATE: $PYTHON db_sqlite/runsql.py db/sqlite/stage/truncate_tables.sql $SCHEMA_STAGE"
	$PYTHON $GXD_PDK_DIR/bin/db_sqlite/runsql.py $GXD_PDK_DIR/db/sqlite/stage/truncate_tables.sql $SCHEMA_STAGE
	if [ $? != 0 ]; then
		msg_err "UPDATE: ERROR: $PYTHON db_sqlite/runsql.py db/sqlite/stage/truncate_tables.sql $SCHEMA_STAGE returned error"
		exit_abnormal
	fi
fi

# Once all steps ran successfully, clear previously failed step
rm -f $VARS/job_phase_step.var

echo
