#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Downloads XML files from the given API endpoint
#  Obtains update_ids from XML using extract_xpath.py script
#  Refreshes current update_id and persists it in vars directory
#  Repeats the process until the end of stream is reached
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

VARS=$GXD_PDK_DIR/vars

# The following variables have been set via loading config:
# BASE_URL, APIKEY, UPDATE_ID_LOCATION, STREAM_ID (optional)
# Now set variables passed via parameters
ENDPOINT=$1
URL_PATH=$2
UPDATE_ID=$3
LIMIT=$4

if [ $UPDATE_ID_LOCATION = "HEADER" ]; then
	NEXT_UPDATE_ID_XPATH=$5
	MAX_UPDATE_ID_XPATH="$6"
elif [ $UPDATE_ID_LOCATION = "BODY" ]; then
	UPDATE_ID_XPATH=$5
fi

API_CALLS=0

while true
do
	if [ -z $STREAM_ID ]; then
		URL="${BASE_URL}/${URL_PATH}?updateId=${UPDATE_ID}&limit=${LIMIT}&api_key=${APIKEY}"
	else
		URL="${BASE_URL}/${URL_PATH}?updateId=${UPDATE_ID}&limit=${LIMIT}&api_key=${APIKEY}&streamId=${STREAM_ID}"
	fi
	FILENAME="files/xml/${ENDPOINT}-${UPDATE_ID}.xml"
	HTTP_STATUS=0
	XML_ERROR=0

	# Fetch endpoint using current update_id; retry a couple of times if HTTP or (very unlikely) XML format errors are encountered
	for ATTEMPT in 1 2 3; do
		msg_ok "DOWNLOAD(API): curl $URL -o $FILENAME"

		HTTP_STATUS=$(curl -w %{http_code} "$URL" -o "$GXD_PDK_DIR/$FILENAME")
		CURL_STATUS=$?

		if [ $CURL_STATUS = 0 ]; then
			if [ $HTTP_STATUS = 200 ]; then
				xmllint --noout --format "$GXD_PDK_DIR/$FILENAME" 2>/dev/null
				XML_ERROR=$?
				if [ $XML_ERROR = 0 ]; then
					# Downloaded a valid XML file, proceed
					break
				fi
			fi
		fi
		# If no valid file for any reason, retry after some time
		msg_ok "DOWNLOAD(API): ERROR: curl exit code: $CURL_STATUS; HTTP response: $HTTP_STATUS; xmllint exit code: $XML_ERROR; attempt: $ATTEMPT. Retrying (up to 3 times) or exiting in 5sec .."
		echo
		sleep 5
	done

	# If curl returned an error
	if [ $CURL_STATUS != 0 ]; then
		msg_err "DOWNLOAD(API): ERROR: curl returned status ${CURL_STATUS} downloading ${URL} to ${FILENAME}"
		rm -f $GXD_PDK_DIR/$FILENAME
		echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		exit_abnormal
	fi

	# If HTTP errors were encountered after multiple attempts:
	if [ $HTTP_STATUS != 200 ]; then
		msg_err "DOWNLOAD(API): ERROR: curl returned HTTP ${HTTP_STATUS} after several attempts fetching ${URL}"
		rm -f $GXD_PDK_DIR/$FILENAME
		echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		exit_abnormal
	fi

	# If XML did not pass format validation:
	if [ $XML_ERROR != 0 ]; then
		msg_err "DOWNLOAD(API): ERROR: xmllint returned status ${XML_ERROR} on ${FILENAME} from URL ${URL}"
		mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
		echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		exit_abnormal
	fi

	# At this point we have parseable XML response obtained using current update_id

	# For APIs with next/max update_id values in the response header (ON, GVD):
	if [ $UPDATE_ID_LOCATION = "HEADER" ]; then

		# Retrieve next update_id
		msg_ok "DOWNLOAD(API): $PYTHON bin/extract_xpath.py $FILENAME $NEXT_UPDATE_ID_XPATH MAX"
		NEXT_UPDATE_ID=$($PYTHON  $GXD_PDK_DIR/bin/extract_xpath.py  $GXD_PDK_DIR/$FILENAME  $NEXT_UPDATE_ID_XPATH  MAX )
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(API): ERROR: extract_xpath.py returned error on ${FILENAME}, ${NEXT_UPDATE_ID_XPATH}"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
			exit_abnormal
		fi
		if [ -z $NEXT_UPDATE_ID ]; then
			NEXT_UPDATE_ID="-"
		fi

		# Retrieve max update_id
		msg_ok "DOWNLOAD(API): $PYTHON bin/extract_xpath.py $FILENAME $MAX_UPDATE_ID_XPATH MAX"
		MAX_UPDATE_ID=$($PYTHON  $GXD_PDK_DIR/bin/extract_xpath.py  $GXD_PDK_DIR/$FILENAME  $MAX_UPDATE_ID_XPATH  MAX )
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(API): ERROR: extract_xpath.py returned error on ${FILENAME}, ${MAX_UPDATE_ID_XPATH}"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
			exit_abnormal
		fi
		if [ -z $MAX_UPDATE_ID ]; then
			MAX_UPDATE_ID="-"
		fi

		# Output download progress indicator
		msg_ok "DOWNLOAD(API): Next/max update_ids found in $FILENAME: $NEXT_UPDATE_ID/$MAX_UPDATE_ID "

		# Check end of stream conditions
		if [ $NEXT_UPDATE_ID = "-" ]; then

			if [ $MAX_UPDATE_ID = "-" ]; then
				# If neither update_ids are present, we simply might not have entitlement for the endpoint
				# Alternatively, API may have returned a valid XML indicating some sort of error and not containing any update_id

				msg_ok "DOWNLOAD(API): API key might not have this endpoint entitlement OR valid XML with an error message was returned. Skipping endpoint"
				rm $GXD_PDK_DIR/$FILENAME
				break

			else
				# Reached the end of stream, check if file contains data and recalculate update_id
				if [ $MAX_UPDATE_ID -lt $UPDATE_ID ]; then
					# No updates since last run, remove file
					msg_ok "DOWNLOAD(API): No update_id greater than ${UPDATE_ID} found in ${FILENAME}, deleting it"
					rm $GXD_PDK_DIR/$FILENAME
				else
					# Next update id will be max update_id plus 1
					UPDATE_ID=$(( $MAX_UPDATE_ID + 1 ))
					echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
				fi
				msg_ok "DOWNLOAD(API): Reached the end of stream"
				break
			fi

		else
			# Not end of stream yet, refresh update_id
			UPDATE_ID=$NEXT_UPDATE_ID
			echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		fi


	# For APIs without next/max update_id values in the response header (GSD/TVSD):
	elif [ $UPDATE_ID_LOCATION = "BODY" ]; then

		# Find the largest update_id in the file
		msg_ok "DOWNLOAD(API): $PYTHON bin/extract_xpath.py $FILENAME $UPDATE_ID_XPATH MAX"
		NEW_UPDATE_ID=$($PYTHON  $GXD_PDK_DIR/bin/extract_xpath.py  $GXD_PDK_DIR/$FILENAME  $UPDATE_ID_XPATH  MAX )
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(API): ERROR: extract_xpath.py returned error on ${FILENAME}, ${UPDATE_ID_XPATH}"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
			exit_abnormal
		fi

		if [ -z $NEW_UPDATE_ID ]; then
			# Reached the end of stream, no data in the file
			msg_ok "DOWNLOAD(API): No update_id found in ${FILENAME}, deleting it"
			rm $GXD_PDK_DIR/$FILENAME
			msg_ok "DOWNLOAD(API): Reached the end of the stream"
			break
		else
			# Not end of stream yet, refresh update_id
			UPDATE_ID=$(( $NEW_UPDATE_ID + 1 ))
			echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		fi
	fi

	# Check if API_CALLS_MAX is reached
	API_CALLS=$(( $API_CALLS + 1 ))
	if [ $API_CALLS = "$API_CALLS_MAX" ]; then
		msg_ok "DOWNLOAD(API): Reached maximum API calls: $API_CALLS_MAX"
		break
	fi

done


