#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Parses FTP file list, decides which files to download based on update dates
#  Downloads and validates selected files, stores new update dates
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

VARS=$GXD_PDK_DIR/vars
mkdir -p $GXD_PDK_DIR/files/xml
#echo "JOB_PHASE=DOWNLOAD_FTP" > $VARS/job_phase.var
echo "JOB_PHASE_STEP=FTP" > $VARS/job_phase_step.var

echo

# Check if any of the FTP variables in the config file are zero/empty
if [ -z $FTP_HOST ] || 
   [ -z $FTP_USER ] || 
   [ -z $FTP_PASS ] ||
   [ -z $FTP_PATH ] ||
   [ -z $FILE_CONFIGS ]
then
	msg_err "DOWNLOAD(FTP): ERROR: One or more variables: FTP_HOST | FTP_USER | FTP_PASS | FTP_PATH | FILE_CONFIGS is not set in config"
	exit_abnormal
fi

msg_ok "DOWNLOAD(FTP): Building list of files to download from FTP..."

# Get the full list of files from configured FTP site/path
for ATTEMPT in 1 2 3; do
	msg_ok "DOWNLOAD(FTP): curl -l -u $FTP_USER@$FTP_HOST/$FTP_PATH/"		
	FTP_FILELIST=$( curl -s -l -u $FTP_USER:$FTP_PASS $FTP_HOST/$FTP_PATH/ )
	CURL_ERROR=$?
	if [ $CURL_ERROR = 0 ]; then
		break
	fi
	sleep 5
done
if [ $CURL_ERROR != 0 ]; then
	msg_err "DOWNLOAD(FTP): ERROR: curl returned error fetching file list from $FTP_USER@$FTP_HOST/$FTP_PATH/"
	exit_abnormal
fi

# Clear stored filelists if any
rm -f $VARS/download_list.tsv
touch $VARS/download_list.tsv
rm -f $VARS/download_list_incr.tsv
touch $VARS/download_list_incr.tsv

# Collect filenames to download for each resource
while read -r CONFIG; do

	RESOURCE=$( echo "$CONFIG" | cut -f1 )
	FILENAME_FULL=$( echo "$CONFIG" | cut -f3 )
	FILENAME_INCR=$( echo "$CONFIG" | cut -f4 )

	# Initialize update_date at zero and load any previously stored update_date value for the resource
	UPDATE_DATE=0
	if [ -e $VARS/update_date_$RESOURCE.var ]; then
	    . $VARS/update_date_$RESOURCE.var
	fi

	# We'll be extracting YYYYMMDD from the filenames (full files), finding the latest YYYYMMDD (that's greater than update_date) and storing the corresponding filename
	SED_PATTERN=s/^$FILENAME_FULL\$/\\1/p
	LATEST_YYYYMMDD=0
	LATEST_FILENAME=""
	for FILENAME in $FTP_FILELIST; do
		YYYYMMDD=$( echo $FILENAME | sed -ne $SED_PATTERN )
		if [ -z $YYYYMMDD ]; then
			:	# pattern no match, do nothing
		else
			if  [ $YYYYMMDD -gt $UPDATE_DATE ] &&
				[ $YYYYMMDD -gt $LATEST_YYYYMMDD ]
			then
				LATEST_YYYYMMDD=$YYYYMMDD
				LATEST_FILENAME=$FILENAME
			fi
		fi
	done
	# If we found the latest YYYYMMDD for the full file, store the resource name, date and filename
	if [ $LATEST_YYYYMMDD != 0 ]; then
		echo "$RESOURCE\t$LATEST_YYYYMMDD\t$LATEST_FILENAME" >> $VARS/download_list.tsv
	fi

	# If incrementals exist for the resource, we'll be extracting YYYYMMDD and storing dates+filenames where YYYYMMDD > update_date AND YYYYMMDD > latest date of the full file
	if [ "$FILENAME_INCR" != "" ]; then
		SED_PATTERN=s/^$FILENAME_INCR\$/\\1/p
		for FILENAME in $FTP_FILELIST; do
			YYYYMMDD=$( echo $FILENAME | sed -ne $SED_PATTERN )
			if [ -z $YYYYMMDD ]; then
				:	# pattern no match, do nothing
			else
				if  [ $YYYYMMDD -gt $UPDATE_DATE ] &&
					[ $YYYYMMDD -gt $LATEST_YYYYMMDD ]
				then
					# Store the resource name, date and filename
					echo "$RESOURCE\t$YYYYMMDD\t$FILENAME" >> $VARS/download_list_incr.tsv
				fi
			fi
		done
	fi

done < $GXD_PDK_DIR/config/$FILE_CONFIGS

# Check if any of the files is not empty and if so, proceed; otherwise exit - no updates
if  [ -s $VARS/download_list.tsv ] ||
	[ -s $VARS/download_list_incr.tsv ]
then
	# Sort download lists (must have ascending udpate_date)
	sort -o $VARS/download_list.tsv $VARS/download_list.tsv
	sort -o $VARS/download_list_incr.tsv $VARS/download_list_incr.tsv

	# Debug output
	echo "\n--- FULLS ---"
	cat $VARS/download_list.tsv
	echo "\n--- INCREMENTALS ---"
	cat $VARS/download_list_incr.tsv
	echo

	# Determine core action depending on the requested ingest type and availability of files
	# Retrieve requested ingest type if any
	FILE_INGEST_TYPE="AUTO"
	if [ -e $VARS/file_ingest_type.var ]; then
		. $VARS/file_ingest_type.var
	fi

	# If incremental ingest is requested, proceed with core update even if we have data for full reload (core replace)
	if [ "$FILE_INGEST_TYPE" = "INCR" ]; then
		echo "CORE_ACTION=UPDATE" > $VARS/core_action.var
	else
		# Check whether we have data for full reload (core replace)
		# To check we compare how many distinct resources are present in the download list (full files) - vs in FILE_CONFIGS
		if [ "$( cut -f1 $VARS/download_list.tsv | sort | uniq | wc -l )" = "$( cut -f1 $GXD_PDK_DIR/config/$FILE_CONFIGS | sort | uniq | wc -l )" ]; then
			# Have data for full reload, proceed with core replace
			echo "CORE_ACTION=REPLACE" > $VARS/core_action.var
		else
			# If full ingest is requested, exit with error as we can't do core replace without full set of resources
			if [ "$FILE_INGEST_TYPE" = "FULL" ]; then
				msg_err "DOWNLOAD(FTP): ERROR: Full ingest is requested but number of resources in vars/download_list.tsv doesn't match the same number in $FILE_CONFIGS"
				exit_abnormal
			else
				# Else (auto ingest) - proceed with core update
				echo "CORE_ACTION=UPDATE" > $VARS/core_action.var
			fi		
		fi		
	fi


	# Download full files and refresh update_date as each file has been successfully downloaded and validated
	while read -r LIST_ENTRY; do

		RESOURCE=$( echo "$LIST_ENTRY" | cut -f1 )
		UPDATE_DATE=$( echo "$LIST_ENTRY" | cut -f2 )
		FTP_FILENAME=$( echo "$LIST_ENTRY" | cut -f3 )

		# Otherwise download/rename, unzip, validate the file and store update_id/replace_date for the resource
		FILENAME="files/xml/$RESOURCE-$UPDATE_DATE.xml.gz"
		msg_ok "DOWNLOAD(FTP): Downloading $FTP_FILENAME for $RESOURCE.."
		for ATTEMPT in 1 2 3; do
			msg_ok "DOWNLOAD(FTP): curl -u $FTP_PATH/$FTP_FILENAME -o $FILENAME"
			curl -u $FTP_USER:$FTP_PASS "$FTP_HOST/$FTP_PATH/$FTP_FILENAME" -o $GXD_PDK_DIR/$FILENAME
			CURL_STATUS=$?
			if [ $CURL_STATUS = 0 ]; then
				break
			fi
			sleep 5
		done		
		if [ $CURL_STATUS != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: curl returned error downloading $FTP_USER@$FTP_HOST/$FTP_PATH/$FTP_FILENAME to $FILENAME"
			rm -f $GXD_PDK_DIR/$FILENAME
			exit_abnormal
		fi

		gunzip -f $GXD_PDK_DIR/$FILENAME
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: gunzip returned error unzipping $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		FILENAME="files/xml/$RESOURCE-$UPDATE_DATE.xml"
		xmllint --noout --format --stream $GXD_PDK_DIR/$FILENAME 2>/dev/null
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: xmllint returned error validating $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		# At this point we've successfully downloaded and validated the "seed" file for the resource
		msg_ok "DOWNLOAD(FTP): Storing new update_date for $RESOURCE: $UPDATE_DATE"
		echo "UPDATE_DATE=$UPDATE_DATE" > $VARS/update_date_$RESOURCE.var

	done < $VARS/download_list.tsv
	mv $VARS/download_list.tsv $GXD_LOG_DIR/

	# Download incremental files and refresh update_date as each file has been successfully downloaded and validated
	while read -r LIST_ENTRY; do

		RESOURCE=$( echo "$LIST_ENTRY" | cut -f1 )
		UPDATE_DATE=$( echo "$LIST_ENTRY" | cut -f2 )
		FTP_FILENAME=$( echo "$LIST_ENTRY" | cut -f3 )

		# Otherwise download/rename, unzip, validate the file and store update_id/replace_date for the resource
		FILENAME="files/xml/$RESOURCE-$UPDATE_DATE.xml.gz"
		msg_ok "DOWNLOAD(FTP): Downloading $FTP_FILENAME for $RESOURCE.."
		for ATTEMPT in 1 2 3; do
			msg_ok "DOWNLOAD(FTP): curl -u $FTP_PATH/$FTP_FILENAME -o $FILENAME"
			curl -u $FTP_USER:$FTP_PASS "$FTP_HOST/$FTP_PATH/$FTP_FILENAME" -o $GXD_PDK_DIR/$FILENAME
			CURL_STATUS=$?
			if [ $CURL_STATUS = 0 ]; then
				break
			fi
			sleep 5
		done		
		if [ $CURL_STATUS != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: curl returned error downloading $FTP_USER@$FTP_HOST/$FTP_PATH/$FTP_FILENAME to $FILENAME"
			rm -f $GXD_PDK_DIR/$FILENAME
			exit_abnormal
		fi

		gunzip -f $GXD_PDK_DIR/$FILENAME
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: gunzip returned error unzipping $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		FILENAME="files/xml/$RESOURCE-$UPDATE_DATE.xml"
		xmllint --noout --format --stream $GXD_PDK_DIR/$FILENAME 2>/dev/null
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(FTP): ERROR: xmllint returned error validating $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		# At this point we've successfully downloaded and validated the incremental file for the resource
		msg_ok "DOWNLOAD(FTP): Storing new update_date for $RESOURCE: $UPDATE_DATE"
		echo "UPDATE_DATE=$UPDATE_DATE" > $VARS/update_date_$RESOURCE.var

	done < $VARS/download_list_incr.tsv
	mv $VARS/download_list_incr.tsv $GXD_LOG_DIR/

else
	msg_ok "DOWNLOAD(FTP): No new files are available"

	# Let the main script proceed with update rather than replacing DB with blank data
	echo "CORE_ACTION=UPDATE" > $VARS/core_action.var

	mv $VARS/download_list.tsv $GXD_LOG_DIR/
	mv $VARS/download_list_incr.tsv $GXD_LOG_DIR/
fi

echo

