#!/bin/sh

#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Retrieves and parses seed file manifest, stores new update_ids
#  Downloads, unzips and validates the seed file for each endpoint
#  =============================================================================

. $GXD_PDK_DIR/bin/functions.sh
load_config

VARS=$GXD_PDK_DIR/vars
mkdir -p $GXD_PDK_DIR/files/xml
#echo "JOB_PHASE=DOWNLOAD_SEED" > $VARS/job_phase.var
echo "JOB_PHASE_STEP=SEED" > $VARS/job_phase_step.var

echo

# Check if any of the FTP variables in the config file are zero/empty
if [ -z $FTP_HOST ] || 
   [ -z $FTP_USER ] || 
   [ -z $FTP_PASS ] ||
   [ -z $FTP_PATH ] ||
   [ -z $SEED_MANIFEST_REGEX ]
then
	msg_err "DOWNLOAD(SEED): ERROR: One or more variables: FTP_HOST | FTP_USER | FTP_PASS | FTP_PATH | SEED_MANIFEST_REGEX is not set in config"
	exit_abnormal
fi

msg_ok "DOWNLOAD(SEED): Checking if new seed manifest is available"

# Get the list of files from configured FTP site/path
for ATTEMPT in 1 2 3; do
	msg_ok "DOWNLOAD(SEED): curl -l -u $FTP_USER@$FTP_HOST/$FTP_PATH/"		
	FILELIST=$( curl -s -l -u $FTP_USER:$FTP_PASS $FTP_HOST/$FTP_PATH/ )
	CURL_ERROR=$?
	if [ $CURL_ERROR = 0 ]; then
		break
	fi
	sleep 5
done
if [ $CURL_ERROR != 0 ]; then
	msg_err "DOWNLOAD(SEED): ERROR: curl returned error fetching file list from $FTP_USER@$FTP_HOST/$FTP_PATH/"
	exit_abnormal
fi

# Find latest YYYYMMDD in the manifest file names
# Use pattern for sed to return YYYYMMDD for the manifest file name
# Empty string will be returned in case of no match
SED_PATTERN=s/^$SEED_MANIFEST_REGEX\$/\\1/p
LATEST_YYYYMMDD=0
LATEST_MANIFEST=""

for FILENAME in $FILELIST; do
	YYYYMMDD=$( echo $FILENAME | sed -ne $SED_PATTERN )
	# Error handling
	if [ -z $YYYYMMDD ]; then
		:	# pattern no match, do nothing
	else
		# NB: not posix-comliant
		if [ $YYYYMMDD \> $LATEST_YYYYMMDD ]; then
			LATEST_YYYYMMDD=$YYYYMMDD
			LATEST_MANIFEST=$FILENAME
		fi
	fi
done

if [ $LATEST_YYYYMMDD = 0 ]; then
	# Found no files matching the pattern, this could indicate an issue
	msg_err "DOWNLOAD(SEED): ERROR: No files found matching the manifest name pattern"
	exit_abnormal
fi

msg_ok "DOWNLOAD(SEED): Most recent YYYYMMDD: $LATEST_YYYYMMDD"

# Initialize replace_date at zero and load previously stored replace_date value if any
REPLACE_DATE=0
if [ -e $VARS/replace_date.var ]; then
    . $VARS/replace_date.var
fi

# If latest YYYYMMDD is greater than our current replace_date, download and parse manifest
# NB: not posix-comliant
if [ $LATEST_YYYYMMDD \> $REPLACE_DATE ]; then

	for ATTEMPT in 1 2 3; do
		msg_ok "DOWNLOAD(SEED): curl -u $FTP_USER@$FTP_HOST/$FTP_PATH/$LATEST_MANIFEST -o vars/seed_manifest.xml"		
		curl -u $FTP_USER:$FTP_PASS "$FTP_HOST/$FTP_PATH/$LATEST_MANIFEST" -o $VARS/seed_manifest.xml
		CURL_ERROR=$?
		if [ $CURL_ERROR = 0 ]; then
			xmllint --noout --format $VARS/seed_manifest.xml 2>/dev/null
			XML_ERROR=$?
			if [ $XML_ERROR = 0 ]; then
				# Downloaded a valid XML manifest, proceed
				break
			fi			
		fi
		sleep 5
	done
	if [ $CURL_ERROR != 0 ]; then
		msg_err "DOWNLOAD(SEED): ERROR: curl returned error fetching $FTP_USER@$FTP_HOST/$FTP_PATH/$LATEST_MANIFEST"
		exit_abnormal
	fi
	if [ $XML_ERROR != 0 ]; then
		msg_err "DOWNLOAD(SEED): ERROR: xmllint returned status ${XML_ERROR} on $VARS/seed_manifest.xml"
		mv $VARS/seed_manifest.xml $GXD_LOG_DIR/errors/
		exit_abnormal
	fi

	msg_ok "DOWNLOAD(SEED): Parsing vars/seed_manifest.xml and extracting data columns.."

	# Extract xpaths from manifest into tsv columns
	XPATH_EXTRACT_ERROR=0
	if [ $XPATH_EXTRACT_ERROR = 0 ]; then
		python3 $GXD_PDK_DIR/bin/extract_xpath.py $VARS/seed_manifest.xml ManifestFile/files/file/endpoint > $VARS/seed_endpoints.tsv
		XPATH_EXTRACT_ERROR=$?
	fi
	if [ $XPATH_EXTRACT_ERROR = 0 ]; then
		python3 $GXD_PDK_DIR/bin/extract_xpath.py $VARS/seed_manifest.xml ManifestFile/files/file/name > $VARS/seed_files.tsv
		XPATH_EXTRACT_ERROR=$?
	fi
	if [ $XPATH_EXTRACT_ERROR = 0 ]; then
		python3 $GXD_PDK_DIR/bin/extract_xpath.py $VARS/seed_manifest.xml ManifestFile/files/file/totalRecordCount > $VARS/seed_counts.tsv
		XPATH_EXTRACT_ERROR=$?
	fi
	if [ $XPATH_EXTRACT_ERROR = 0 ]; then
		python3 $GXD_PDK_DIR/bin/extract_xpath.py $VARS/seed_manifest.xml ManifestFile/files/file/lastUpdateId > $VARS/seed_updateids.tsv
		XPATH_EXTRACT_ERROR=$?
	fi
	if [ $XPATH_EXTRACT_ERROR != 0 ]; then
		msg_err "DOWNLOAD(SEED): ERROR: bin/extract_xpath.py returned error extracting one of xpaths from vars/seed_manifest.xml"
		mv $VARS/seed_manifest.xml $GXD_LOG_DIR/errors/
		exit_abnormal
	fi

	# We've now extracted columns of data from manifest XML
	# We will combine them into seed_manifest.tsv file using paste command
	# This relies on all datapoints being present for all endpoints in the manifest XML file
	# As a precaution, we'll check that we have the same number of entries in each extracted file
	MANIFEST_ENDPOINT_COUNT=$( wc -l < $VARS/seed_endpoints.tsv )
	if 	[ $MANIFEST_ENDPOINT_COUNT = 0 ] ||
		[ $MANIFEST_ENDPOINT_COUNT != $( wc -l < $VARS/seed_files.tsv ) ] ||
		[ $MANIFEST_ENDPOINT_COUNT != $( wc -l < $VARS/seed_counts.tsv ) ] ||
		[ $MANIFEST_ENDPOINT_COUNT != $( wc -l < $VARS/seed_updateids.tsv ) ]
	then
		msg_err "DOWNLOAD(SEED): ERROR: one or more elements (endpoint, file name, count, update id) are missing from manifest"
		mv $VARS/seed_manifest.xml $GXD_LOG_DIR/errors/
		exit_abnormal
	fi

	# Combine columns of data from manifest XML into the TSV file
	msg_ok "DOWNLOAD(SEED): Merging extracted data columns into single vars/seed_manifest.tsv.."
	paste $VARS/seed_endpoints.tsv $VARS/seed_files.tsv $VARS/seed_counts.tsv $VARS/seed_updateids.tsv > $VARS/seed_manifest.tsv

	while read -r MANIFEST_ROW; do

		ENDPOINT=$( echo "$MANIFEST_ROW" | cut -f1 )
		SEED_FILE=$( echo "$MANIFEST_ROW" | cut -f2 )
		OBJECT_COUNT=$( echo "$MANIFEST_ROW" | cut -f3 )
		LAST_UPDATE_ID=$( echo "$MANIFEST_ROW" | cut -f4 )

		# If the manifest endpoint is not part of configured endpoints, skip to the next one
		if [ "$( cut -f1 $GXD_PDK_DIR/config/$ENDPOINT_CONFIGS | grep -w $ENDPOINT )" = "" ]; then
			continue
		fi

		# Retrieve stored replace date for the endpoint (if any)
		REPLACE_DATE=0
		if [ -e $VARS/replace_date_$ENDPOINT.var ]; then
		    . $VARS/replace_date_$ENDPOINT.var
		fi

		# If we've already processed the endpoint and stored the latest replace date, skip it
		if [ $REPLACE_DATE = $LATEST_YYYYMMDD ]; then
			continue
		fi

		echo

		# Otherwise download/rename, unzip, validate the file and store update_id/replace_date for endpoint
		FILENAME="files/xml/$ENDPOINT-0.xml.gz"
		msg_ok "DOWNLOAD(SEED): Downloading seed file for $ENDPOINT.."
		for ATTEMPT in 1 2 3; do
			msg_ok "DOWNLOAD(SEED): curl -u $FTP_PATH/$SEED_FILE.gz -o $FILENAME"
			curl -u $FTP_USER:$FTP_PASS "$FTP_HOST/$FTP_PATH/$SEED_FILE.gz" -o $GXD_PDK_DIR/$FILENAME
			CURL_STATUS=$?
			if [ $CURL_STATUS = 0 ]; then
				break
			fi
			sleep 5
		done		
		if [ $CURL_STATUS != 0 ]; then
			msg_err "DOWNLOAD(SEED): ERROR: curl returned error downloading $FTP_USER@$FTP_HOST/$FTP_PATH/$SEED_FILE.gz to $FILENAME"
			rm -f $GXD_PDK_DIR/$FILENAME
			exit_abnormal
		fi

		gunzip -f $GXD_PDK_DIR/$FILENAME
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(SEED): ERROR: gunzip returned error unzipping $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		FILENAME="files/xml/$ENDPOINT-0.xml"
		xmllint --noout --format --stream $GXD_PDK_DIR/$FILENAME 2>/dev/null
		if [ $? != 0 ]; then
			msg_err "DOWNLOAD(SEED): ERROR: xmllint returned error validating $FILENAME"
			mv $GXD_PDK_DIR/$FILENAME $GXD_LOG_DIR/errors/
			exit_abnormal
		fi

		# At this point we've successfully downloaded and validated the seed file for the endpoint
		UPDATE_ID=$(( LAST_UPDATE_ID + 1 ))
		msg_ok "DOWNLOAD(SEED): Storing new update_id for endpoint $ENDPOINT: $UPDATE_ID, replace_date: $LATEST_YYYYMMDD"
		echo "UPDATE_ID=$UPDATE_ID" > $VARS/update_id_$ENDPOINT.var
		echo "REPLACE_DATE=$LATEST_YYYYMMDD" > $VARS/replace_date_$ENDPOINT.var

	done < $VARS/seed_manifest.tsv

	# At this point we successfully downloaded and validated all available seed files for all endpoints so we can update replace date
	echo "REPLACE_DATE=$LATEST_YYYYMMDD" > $VARS/replace_date.var

	rm $VARS/seed_*.xml
	rm $VARS/seed_endpoints.tsv
	rm $VARS/seed_files.tsv
	rm $VARS/seed_counts.tsv
	rm $VARS/seed_updateids.tsv

else
	msg_ok "DOWNLOAD(SEED): No new seed manifest available"
	# Run script will compare previous replace_date with new replace_date to see if it needs to proceed with API download
fi

echo
