#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Common Python functions used by various components of the product demo kit
#  =============================================================================

import os
import io
import sys
import csv
import json
import hashlib
import datetime
import xml.etree.cElementTree as ET

is_py2 = sys.version_info[0] == 2
if is_py2:
	import contextlib2 as contextlib
else:
	import contextlib


# Logging/output

def log(message):
	with open(os.environ['GXD_LOG'], "a") as log_file:
		log_file.write(message)

def ok(message):
	sys.stdout.write(message)
	log(message)

def error(message):
	sys.stderr.write(message)
	log(message)


# Type checks

def is_int(input, check_trim=False):
	if check_trim and input.strip() != input:
		return False	
	try:
		num = int(input)
	except ValueError:
		return False
	return True

def is_float(input, check_trim=False):
	if check_trim and input.strip() != input:
		return False	
	try:
		num = float(input)
	except ValueError:
		return False
	return True

def is_datetime(input, format = '%Y-%m-%dT%H:%M:%SZ', check_trim=False):
	if check_trim and input.strip() != input:
		return False	
	try:
		dt = datetime.datetime.strptime(input, format)
	except ValueError:
		return False
	return True


# Generic utility functions

def index(list_of_names):
	idx = {}
	for i, name in enumerate(list_of_names):
		idx[name] = i
	return idx

def remove_none_values (dict_or_list):
	if isinstance(dict_or_list, dict):
		for key, value in list(dict_or_list.items()):
			if value is None:
				del dict_or_list[key]
			else:
				remove_none_values(value)
	if isinstance(dict_or_list, list):
		for value in dict_or_list:
			remove_none_values(value)
	return dict_or_list

def group_iterator(iterator, group_size):
	group = []
	for item in iterator:
		group.append(item)
		if len(group) == group_size:
			yield group
			group = []
	if len(group) > 0:
		yield group


# PDK config/fields utility functions

def read_config(config_file):
	config = {}
	if os.path.isfile(config_file):
		with open(config_file) as cfg:
			for line in cfg:
				if line.strip() == '' or line.startswith('#'):
					continue

				line 	= line.strip('\n')
				key 	= line.split("=")[0]
				value 	= "=".join(line.split("=")[1:])
				
				# Remove leading/terminating single quotes from strings (passwords)
				if value.startswith("'") and value.endswith("'"):
					value = value[1:-1]

				if str(value) == "''" or str(value) == '""':
					value = ''

				if str(value).upper() == 'TRUE':
					value = True
				elif str(value).upper() == 'FALSE':
					value = False
				else:
					pass

				config[key] = value
	return config

def get_config_setting(setting, pdk_dir, custom_conf):
	default = read_config( os.path.join(pdk_dir, "config", "default.conf") )
	custom 	= read_config( os.path.join(pdk_dir, "config", custom_conf) )
	value 	= custom.get(setting) if custom.get(setting) is not None else default.get(setting)
	return value

def get_fields(fields_tsv_filename, table_name):
	fields = {}

	# fields.tsv columns enum
	Entity, Table, Ordinal, Name, Type, Nullable, MaxLength, PdkProperties, Xpath, Shell, Stats = range(0, 11)

	with open(fields_tsv_filename) as fields_tsv_file:
		index = 0
		for row in csv.reader(fields_tsv_file, delimiter = "\t"):
			if table_name == row[ Table ]:
				name = row[ Name ]

				props = {}
				props['ord'] = row[ Ordinal ]
				props['name'] = name
				props['type'] = row[ Type ]
				props['is_nullable'] = True if row[ Nullable ].upper() == 'TRUE' else False
				props['max_length'] = None if row[ MaxLength ].upper() == 'NONE' else int(row[ MaxLength ])
				props['idx'] = index
				props['table'] = table_name

				pdk_field_primary = row[ PdkProperties ].split('|')[0]
				props['pdk_field_type'] = None if pdk_field_primary.upper() == 'NONE' else pdk_field_primary.split(':')[0]
				props['pdk_field_value'] = None if len(pdk_field_primary.split(':')) < 2 else pdk_field_primary.split(':')[1]

				pdk_field_properties = []
				for kv in row[ PdkProperties ].split('|'):
					if len(kv.split(':')) < 2:
						pdk_field_properties.append({ kv : None })
					else:
						pdk_field_properties.append({ kv.split(':')[0] : kv.split(':')[1] })

				props['pdk_field_properties'] = pdk_field_properties
				props['shell'] = row [ Shell ]

				fields[name] = props
				index += 1
	return fields

def get_pdk_field_name(fields, pdk_field):
	name_list = [ name for name in fields if fields.get(name).get('pdk_field_type') == pdk_field ]
	return name_list[0] if name_list else None


# Iterators for specific file types

def tsv_iterator(tsv, null_value = '', injections = None ):
	is_py2 = sys.version_info[0] == 2
	with open(tsv) as tsv_handle:
		for row in csv.reader(tsv_handle, delimiter = "\t"):
			if is_py2:
				row = [ None if s == null_value else s.decode("utf8") for s in row]
			else:
				row = [ None if s == null_value else s for s in row]

			# Inject additional fields
			for i in injections:
				row[ i.get('index') ] = i.get('function')()

			yield row

def jsonl_iterator(jsonl, injections = None ):
	with open(jsonl) as jsonl_handle:
		for line in jsonl_handle:
			document = json.loads(line)

			# Inject additional fields
			for i in injections:
				document[ i.get('key') ] = i.get('function')()

			yield document

def sql_iterator(sql_file, separator = ';'):
	# using io.open() to be able to read from stdin (0)
	with io.open(sql_file) as sql:
		statement = []
		for line in sql:
			if line.strip() == '' or line.startswith('--'):
				continue
			statement.append(line.strip())
			if line.strip().endswith(separator):
				if len(statement) > 0:
					yield " ".join(statement)
					statement = []


# Parser parent class


class GnEndpointParser(object):


	def __init__(self, resource, entities, tables, field_names):
		self.resource 	 = resource
		self.entities	 = entities
		self.tables 	 = tables
		self.field_names = field_names
		self.indices 	 = { table: index(field_list) for table, field_list in self.field_names.items() }
		self.is_py2 	 = sys.version_info[0] == 2

	# Init configuration
	def set_config(self, pdk_dir, pdk_custom_config, files_dir):
		self.files_dir = files_dir

		fields_tsv = os.path.join(pdk_dir, "config", get_config_setting('FIELD_CONFIGS', pdk_dir, pdk_custom_config))

		self.fields 	= { table: get_fields( fields_tsv, table ) for table in self.tables }
		self.md5s  		= { table: get_pdk_field_name( fields, 'ROW_HASH' ) for table, fields in self.fields.items() }
		self.resources  = { table: get_pdk_field_name( fields, 'RESOURCE' ) for table, fields in self.fields.items() }

		# Set config parameters
		self.null_value 	= get_config_setting('NULL_VALUE', pdk_dir, pdk_custom_config)
		self.do_validate 	= get_config_setting('VALIDATE', pdk_dir, pdk_custom_config)
		self.trim_check		= get_config_setting('NONTEXT_TRIM_CHECK', pdk_dir, pdk_custom_config)
		self.bool_to_int 	= get_config_setting('BOOL_TO_INT', pdk_dir, pdk_custom_config)
		self.hash_orig_data	= get_config_setting('HASH_ORIG_DATA', pdk_dir, pdk_custom_config)
		self.rdb_data_load 	= get_config_setting('RDB_DATA_LOAD', pdk_dir, pdk_custom_config)
		self.rdb_skeleton 	= False if self.rdb_data_load.upper() == 'FULL' else True

		# Init parser options from config file
		# Usage from parser: str = self.parser_options.get(option)
		self.parser_options = {}
		parser_options = get_config_setting('PARSER_OPTIONS', pdk_dir, pdk_custom_config)
		if parser_options:
			for po in parser_options.split('|'):
				option = po.split(':')
				self.parser_options[ option[0] ] = None if len(option) == 1 else option[1]


		# Initialize skeleton flags for each table/field in the resource

		# Determine if field should be a skeleton field
		def is_skeleton(field):
			if (field['pdk_field_type'] in ['PRIMARY_KEY', 'UPDATE_ID', 'UPDATE_DATE', 'DELETE_FLAG', 'INACTIVE', 'PARENT', 'PARENT_TABLE', 'ENTITY_MAIN_REF', 'ENTITY_DEP_REF'] or
					self.rdb_data_load in str(field['shell']).split(",")):
				return True
			else:
				return False


		# By default, we'd like to only load parent entity tables in skeleton mode, however we may be instructed to include certain child tables as well
		included_child_tables = []

		for table, fields in self.fields.items():

			# entity parent table -> rows are always inserted
			if not get_pdk_field_name(fields, 'PARENT'):

				for name, field in fields.items():
					if is_skeleton(field):
						field['is_skeleton'] = True
					else:
						field['is_skeleton'] = False

			# entity child table -> we want to skip child table rows entirely unless some of the data fields are designated skeleton
			else:

				# determine whether child table has any fields to be included
				# We're omitting ROW_HASH so that "skeleton" child table can be included by labeling it
				child_table_exclusion_fields = ['PRIMARY_KEY', 'PARENT', 'PARENT_TABLE', 'UPDATE_ID', 'INGEST_ID']

				for name, field in fields.items():
					field['is_skeleton'] = False
					if field['pdk_field_type'] not in child_table_exclusion_fields:
						if table not in included_child_tables and is_skeleton(field):
							included_child_tables.append(table)

		# Recursive function to set skeleton field flag on a given table and its immediate parent table
		def include_child_table(table):
			fields = self.fields[table]
			for name, field in fields.items():
				if is_skeleton(field):
					field['is_skeleton'] = True

			parent_table_fk = get_pdk_field_name(fields, 'PARENT_TABLE')
			if parent_table_fk is not None:
				include_child_table( fields[parent_table_fk]['pdk_field_value'] )
			else:
				return

		# Recursively include specified child tables and their immediate parent tables
		for table in included_child_tables:
			include_child_table(table)		
	

		# Set type/format of the parsing output files
		database = get_config_setting('DATABASE', pdk_dir, pdk_custom_config)
		if database in ['MONGO','NONE','NOSQL']:
			self.do_output_jsonl = True
		else:
			self.do_output_jsonl = False
		if database in ['SQLITE','PGSQL','MYSQL','NONE','RELDB']:
			self.do_output_tsv = True
		else:
			self.do_output_tsv = False

		# Flag to skip unessential fields (as determined by parser code, not by skeleton field definitions)
		# Typically will be False only to speed up parsing in skeleton-only RDB ingest, otherwise True
		self.do_parse_all = True

		# SQLite: values of fields declared as Boolean must be integers (as of this writing)
		# The following will force the value conversion in TSV validator function even if the config flag says otherwise
		# This eliminates having to update bool_to_int config flag when switching DB engines
		if database == 'SQLITE' and not self.bool_to_int:
			self.sqlite_force_bool_to_int = True
		else:
			self.sqlite_force_bool_to_int = False


	# Checking XML node or attribute value for nulls
	def check(self, node, attribute = None):
		if node is None:
			return None
		if attribute is None:
			if node.text is None:
				return ''
			else:
				return node.text
		else:
			return node.get(attribute)


	def format_boolean(self, value):
		if value is None:
			return None
		if self.bool_to_int:
			if value.upper() == 'TRUE':
				value = '1'
			elif value.upper() == 'FALSE':
				value = '0'
		return value


	def format_datetime(self, value, input_format = '%Y-%m-%dT%H:%M:%SZ'):
		if value is None:
			return None

		# Python < 3.2 does not support timezone offset (%z) in strptime format
		if '%z' in input_format and sys.version_info[:2] < (3,2):
			return None

		if is_datetime(value, input_format):
			dt = datetime.datetime.strptime(value, input_format)

			# Work around Python 2.7 strftime year restriction (must be >=1900)
			if dt.year < 1900 and sys.version_info[0] == 2:
				return None

			return dt.strftime('%Y-%m-%dT%H:%M:%SZ')
		else:
			return None

	def md5(self, s):
		return hashlib.md5(s.encode('utf-8')).hexdigest()

	def xml_nodes(self, filename, element_xpath, do_skip_root = True, on_start = False):
		path_array = element_xpath.split('/')
		tag = path_array[-1]
		tag_level = len(path_array)
		path_start = 1 if do_skip_root else 0

		# optionally drop first element if skipping/ignoring root (e.g. to avoid dealing with root element namespaces)
		tag_path = '/'.join( path_array[ path_start : tag_level ] )

		level = 0
		path = []

		in_target_element = False
		for event, element in ET.iterparse(filename, events = ('end','start')):
			if event == 'start':
				level += 1
				path.append(element.tag)
				if level == tag_level and element.tag == tag and tag_path == '/'.join( path[ path_start : tag_level ] ):
					in_target_element = True
					if on_start:
						yield element
			elif event == 'end':
				if level == tag_level and element.tag == tag and tag_path == '/'.join( path[ path_start : tag_level ] ):
					if not on_start:
						yield element
					in_target_element = False
				if not in_target_element:
						element.clear()
				level -= 1
				path.pop()


	# Write document to JSONL
	def write_jsonl(self, document, entity):
		document = remove_none_values(document)
		try:
			self.jsonl_handles[entity].write(json.dumps(document) + "\n")
		except IOError as e:
			error("\t{0}: Exception parsing {1} into JSONL: {2}".format( os.path.basename(__file__), self.file_to_parse, str(e) ))
			sys.exit(1)	


	# Validate and write row to TSV
	def validate_and_write_tsv(self, data, table, locator_field):

		idx 		= self.indices[table]
		writer 		= self.tsv_writers[table]
		fields 		= self.fields[table]
		hash_field 	= self.md5s[table]
		rsrc_field	= self.resources[table]


		have_skeleton_fields = False
		row = [ None ] * len(idx)
		row_orig = [ None ] * len(idx)
		for key, value in data.items():

			# Only write data for fields present in relational schema (fields.tsv)
			if idx.get(key) is not None:

				# Preserve original row for MD5 calculation (sans update info)
				if fields[key]['pdk_field_type'] not in ['UPDATE_ID', 'UPDATE_DATE']:
					row_orig[ idx[key] ] = value

				if self.rdb_skeleton and not fields[key]['is_skeleton']:
					continue
				else:
					row[ idx[key] ] = value
					have_skeleton_fields = True

		# Skip write and exit if we're in RDB skeleton mode and table has no fields to write
		if self.rdb_skeleton and not have_skeleton_fields:
			return False

		if self.do_validate:
			result = self.validate_row(row, fields, locator_field = locator_field)

			if not result['is_row_valid']:
				print(result['validation_msg'])
			elif result['is_row_updated']:
				row = result['row']
				if result['validation_msg'] != '':
					print(result['validation_msg'])
			elif result['validation_msg'] != '':
					print(result['validation_msg'])

		if self.hash_orig_data:
			# Use original row (sans update info) before skeleton pruning & validation corrections to calculate MD5
			row_hash = hashlib.md5(str(row_orig).encode('utf-8')).hexdigest()
		else:
			row_hash = hashlib.md5(str(row).encode('utf-8')).hexdigest()
		row[ fields.get(hash_field).get('idx') ] = row_hash

		if rsrc_field:
			row[ fields.get(rsrc_field).get('idx') ] = self.resource

		# Update with configured null value
		row = [ self.null_value if field is None else field for field in row]

		try:
			if not self.do_validate or result['is_row_valid']:
				if self.is_py2:
					writer.writerow([field.encode("utf-8") for field in row])
				else:
					writer.writerow([field for field in row])
				return True
			else:
				return False
		except IOError as e:
			error("\t{0}: Exception parsing {1} into TSV: {2}".format( os.path.basename(__file__), self.file_to_parse, str(e) ))
			sys.exit(1)


	# Validate each field in the row to be written against field definitions (fields.tsv). Only applicable to RDBMS/TSV
	def validate_row(self, row, fields, locator_field = None, row_num = None, null_value = None):

		validation_msg = ""
		is_row_valid = True
		is_row_updated = False

		# Get some identifying label for debugging of rows failing validation - either a specific field value or passed row_num
		locator = str(row_num) if locator_field is None else "{}:{}".format( locator_field, row[ fields.get(locator_field).get('idx') ] )

		# Fields that don't need validation
		skip_validation = [ get_pdk_field_name(fields, pdk_field) for pdk_field in ['INGEST_ID', 'BATCH_UPDATE_ORD'] ]

		# Determine if the row is a shell record being deleted (schema-designated delete flag is True)
		delete_flag_field = get_pdk_field_name(fields, 'DELETE_FLAG')
		if delete_flag_field:
			delete_flag_value = row[ fields.get(delete_flag_field).get('idx') ]
			if delete_flag_value == '1' or delete_flag_value == 'TRUE':
				is_row_being_deleted = True	
			else:
				is_row_being_deleted = False
		else:
			is_row_being_deleted = False

		for name, field in fields.items():

			if name in skip_validation:
				continue

			idx = field['idx']
			table = field['table']
			value = row[idx]

			if value != null_value:

				# Type check			
				if field['type']=='VARCHAR':
					pass

				elif field['type']=='INT' or \
					 field['type']=='BIGINT' or \
					 field['type']=='SMALLINT':
					if not is_int(value, check_trim=self.trim_check):
						value = null_value
						is_row_updated = True
						validation_msg += "Data validation warning - {}:{} - non-integer field ({}), replacing with NULL; ".format( table, locator, name )

				elif field['type']=='DATETIME':
					if not is_datetime(value, check_trim=self.trim_check):
						value = null_value
						is_row_updated = True
						validation_msg += "Data validation warning - {}:{} - non-datetime field ({}), replacing with NULL; ".format( table, locator, name )
					else:
						value = "{}".format(value.replace("T"," ").replace("Z",""))

				elif field['type']=='FLOAT':
					if not is_float(value, check_trim=self.trim_check):
						value = null_value
						is_row_updated = True
						validation_msg += "Data validation warning - {}:{} - non-float field ({}), replacing with NULL; ".format( table, locator, name )

				elif field['type']=='BOOL':
					# Forced replacement of True/False -> 1/0 for SQLite
					if self.sqlite_force_bool_to_int:
						if value.upper() == 'TRUE':
							value = '1'
						elif value.upper() == 'FALSE':
							value = '0'
						is_row_updated = True

				else:
					pass
				
				# Length check
				if field['max_length'] is not None:
					if len(value) > field['max_length']:
						value = value[ :field.get('max_length') ]
						is_row_updated = True
						validation_msg += "Data validation warning - {}:{} - exceeds field length ({}), truncating; ".format( table, locator, name )


				# UpdateId out of bounds check
				if name == get_pdk_field_name(fields, 'UPDATE_ID') and int(value)<int(self.file_update_id):
					validation_msg += "Data validation warning - {}:{} - {} out of bounds ({}<{}); ".format( table, locator, name, value, self.file_update_id )

			# Not nullable check
			if value == null_value and not field['is_nullable']:

				# If row is being deleted, we want it to pass validation (and Staging inserts), so we'll set non-nullable fields to empty values
				# Same approach if skeleton RDB data mode is enabled and the field is non-skeleton (most of them should be nullable anyway)
				if is_row_being_deleted or self.rdb_skeleton and not field['is_skeleton']:

					if field['type']=='VARCHAR' or \
					   field['type']=='TEXT':
						value = ""

					elif field['type']=='INT' or \
						 field['type']=='BIGINT' or \
						 field['type']=='SMALLINT':
						value = 0

					elif field['type']=='DATETIME':
						value = "1900-01-01 00:00:00"

					elif field['type']=='FLOAT':
						value = 0.0

					elif field['type']=='BOOL':
						value = False

					is_row_updated = True

				# If row is not being deleted, it will fail validation and won't be inserted in Staging
				else:				

					is_row_valid = False
					validation_msg += "Data validation error - {}:{} - non-nullable field ({}) is NULL, skipping entire row; ".format( table, locator, name )

			row[idx] = value

		if is_row_valid and not is_row_updated:
			return { 'is_row_valid':is_row_valid, 'is_row_updated':is_row_updated, 'validation_msg':validation_msg }
		else:
			return { 'is_row_valid':is_row_valid, 'is_row_updated':is_row_updated, 'validation_msg':validation_msg, 'row':row }


	# Method to call from each parser script to run parsing
	def run(self, file_to_parse):

		def output_filename(name, update_id, data_type):
			return os.path.join( self.files_dir, data_type, self.resource, "{}-{}.{}".format(name, update_id, data_type) )

		update_id = file_to_parse.split('.')[-2].split('-')[-1]
		self.file_update_id = update_id

		jsonl_fnames 	= [ (entity, output_filename(entity, update_id, 'jsonl')) for entity in self.entities ]
		tsv_fnames		= [ (table, output_filename(table, update_id, 'tsv')) for table in self.tables ]
		
		self.file_to_parse = file_to_parse

		with contextlib.ExitStack() as stack:

			if self.do_output_tsv:
				tsv_handles  		= { table: stack.enter_context(open(filename, 'w')) for table, filename in tsv_fnames }
				self.tsv_writers 	= { table: csv.writer( tsv_handle, delimiter = '\t' ) for table, tsv_handle in tsv_handles.items() }

			if self.do_output_jsonl:
				self.jsonl_handles = { entity: stack.enter_context(open(filename, 'w')) for entity, filename in jsonl_fnames }

			self.parse(file_to_parse)


	# Method contains the actual parsing code and is implemented for each parser
	def parse(self, file_to_parse):
		pass

