#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Parses a given ON Schedules XML file
#  Validates each record against fields configuration file
#  Writes validated records into TSV files for the endpoint
#  =============================================================================

import os
import sys

sys.path.append(os.path.join(sys.path[0],'..'))
import gxd_functions as gxd

xml_file 	= sys.argv[1]
files_path 	= sys.argv[2]

class GnSchedulesParser(gxd.GnEndpointParser):

	def parse(self, xml_file):

		for s in self.xml_nodes(xml_file, "on/schedules/schedule"):

			schedule = {}

			prgSvcId 	= self.check( s, 'prgSvcId' )
			updateId 	= self.check( s, 'updateId' )
			updateDate 	= self.check( s, 'updateDate' )
			date 		= self.check( s, 'date' )
			scheduleId  = "{0}-{1}".format( str(prgSvcId), str(date) )
			date 		= self.format_datetime( date, input_format = '%Y-%m-%d' )

			schedule['scheduleId'] 		= scheduleId
			schedule['updateId'] 		= updateId
			schedule['updateDate'] 		= updateDate
			schedule['prgSvcId'] 		= prgSvcId
			schedule['date'] 			= date

			if self.do_output_tsv:
				if not self.validate_and_write_tsv(schedule, 'schedule', 'scheduleId'):
					s.clear()
					continue

			for se in s.findall('event'):

				event = {}

				TMSId 	= self.check( se, 'TMSId' )
				time 	= self.check( se, 'time' )
				dur 	= self.check( se, 'dur' )
				eventId = "{0}-{1}-{2}-{3}".format( scheduleId, time, dur, TMSId )

				event['eventId'] 	= eventId
				event['scheduleId'] = scheduleId
				event['updateId'] 	= updateId
				event['TMSId'] 		= TMSId
				event['time'] 		= time
				event['dur'] 		= dur

				event['gnEventId'] 		= self.check( se, 'eventId' )
				event['rootId'] 		= self.check( se, 'rootId' )
				event['exactStart'] 	= self.check( se, 'exactStart' )
				event['exactDur'] 		= self.check( se, 'exactDur' )
				event['exactSource'] 	= self.check( se, 'exactSource' )

				event['exactStartDateTime'] = self.format_datetime( self.check( se, 'exactStartDateTime' ))
				event['exactEndDateTime'] 	= self.format_datetime( self.check( se, 'exactEndDateTime' ))

				event['subtitled'] 		= self.format_boolean( self.check( se.find('tv/subtitled') ))
				event['subtitled_lang'] = self.check( se.find('tv/subtitled'), 'lang' )
				event['dubbed'] 		= self.format_boolean( self.check( se.find('tv/dubbed') ))
				event['dubbed_lang'] 	= self.check( se.find('tv/dubbed'), 'lang' )
				event['sap'] 			= self.format_boolean( self.check( se.find('tv/sap') ))
				event['sap_lang'] 		= self.check( se.find('tv/sap'), 'lang')

				event['tvRating'] 		= self.check( se.find('tv/tvRating') )
				event['tvRating_body'] 	= self.check( se.find('tv/tvRating'), 'body' )

				event['tvSubRating'] 		= self.check( se.find('tv/tvSubRating') )
				event['tvSubRating_body'] 	= self.check( se.find('tv/tvSubRating'), 'body' )

				event['partNum'] 		= self.check( se.find('tv/partNum') )
				event['numOfParts'] 	= self.check( se.find('tv/partNum'), 'numOfParts' )

				event['netSynSource'] 	= self.check( se.find('tv/netSynSource') )
				event['netSynType'] 	= self.check( se.find('tv/netSynType') )

				event['quals'] 			= self.check( se.find('quals') )

				event['catchup_validFrom'] 	= self.format_datetime( self.check( se.find('catchup/validFrom') ))
				event['catchup_validTo'] 	= self.format_datetime( self.check( se.find('catchup/validTo') ))

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(event, 'schedule_event', 'eventId'):
						se.clear()
						continue

				for seb in se.findall('broadcast/id'):

					bid = {}

					bid['updateId'] 	= updateId
					bid['scheduleId'] 	= scheduleId
					bid['eventId'] 		= eventId

					bid['broadcastId'] 		= self.check( seb )
					bid['broadcastId_type'] = self.check( seb, 'type' )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(bid, 'schedule_event_bcastid', 'eventId'):
							seb.clear()
							continue

					if self.do_output_jsonl:
						[ bid.pop(key) for key in ['scheduleId','eventId','updateId'] ]
						event['broadcast_ids'] = ( event.get('broadcast_ids') or [] ) + [ bid.copy() ]

					seb.clear()


				right_inc = 0
				for ser in se.findall('contentRights/contentRight'):

					cr = {}
					right_inc += 1

					cr['updateId'] 		= updateId
					cr['scheduleId'] 	= scheduleId
					cr['eventId'] 		= eventId

					cr['customer'] 		= self.check( ser, 'customer' )
					cr['right'] 		= self.check( ser, 'type' )
					cr['allowed'] 		= self.format_boolean( self.check( ser, 'value' ))

					rightId 			= "{0}-{1}-{2}-{3}-{4}".format( eventId, cr['right'], cr['allowed'], cr['customer'], right_inc)
					rightId 			= self.md5(rightId)

					cr['rightId'] 		= rightId

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(cr, 'schedule_event_right', 'rightId'):
							ser.clear()
							continue

					for serp in ser.findall('contentRightParameter'):

						crp = {}

						crp['updateId'] 	= updateId
						crp['scheduleId'] 	= scheduleId
						crp['rightId'] 		= rightId

						crp['param'] 		= self.check( serp )
						crp['type'] 		= self.check( serp, 'type' )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(crp, 'schedule_event_right_param', 'rightId'):
								serp.clear()
								continue

						if self.do_output_jsonl:
							[ crp.pop(key) for key in ['scheduleId','rightId','updateId'] ]
							cr['params'] = ( cr.get('params') or [] ) + [ crp.copy() ]

						serp.clear()


					if self.do_output_jsonl:
						[ cr.pop(key) for key in ['scheduleId','eventId','rightId','updateId'] ]
						event['content_rights'] = ( event.get('content_rights') or [] ) + [ cr.copy() ]

					ser.clear()


				if self.do_output_jsonl:
					[ event.pop(key) for key in ['scheduleId','eventId','updateId'] ]
					schedule['events'] = ( schedule.get('events') or [] ) + [ event.copy() ]

				se.clear()

			if self.do_output_jsonl:
				self.write_jsonl(schedule, 'schedule')

			s.clear()


# ================== ENDPOINT SPECIFIC INIT CODE BEGIN ==================
# Here insert the endpoint-specific code from schema/release/<product>/parse_.py


endpoint = 'Schedules'

entities = [
	 'schedule' 
]

tables = [
	 'schedule' 
	,'schedule_event' 
	,'schedule_event_bcastid' 
	,'schedule_event_right' 
	,'schedule_event_right_param' 
]

fields = {
	 'schedule': ['scheduleId', 'updateId', 'updateDate', 'prgSvcId', 'date', 'MD5', 'ingestId', 'batchUpdateOrd']
	,'schedule_event': ['eventId', 'scheduleId', 'updateId', 'gnEventId', 'TMSId', 'time', 'dur', 'rootId', 'exactStartDateTime', 'exactEndDateTime', 'exactStart', 'exactDur', 'exactSource', 'subtitled', 'subtitled_lang', 'dubbed', 'dubbed_lang', 'sap', 'sap_lang', 'tvRating', 'tvRating_body', 'tvSubRating', 'tvSubRating_body', 'partNum', 'numOfParts', 'netSynSource', 'netSynType', 'quals', 'catchup_validFrom', 'catchup_validTo', 'MD5', 'ingestId']
	,'schedule_event_bcastid': ['eventId', 'scheduleId', 'updateId', 'broadcastId', 'broadcastId_type', 'MD5', 'ingestId']
	,'schedule_event_right': ['rightId', 'eventId', 'scheduleId', 'updateId', 'right', 'allowed', 'customer', 'MD5', 'ingestId']
	,'schedule_event_right_param': ['rightId', 'scheduleId', 'updateId', 'type', 'param', 'MD5', 'ingestId']
}

# =================== ENDPOINT SPECIFIC INIT CODE END ===================


gnSchedulesParser = GnSchedulesParser( endpoint, entities, tables, fields )
gnSchedulesParser.set_config( os.environ['GXD_PDK_DIR'], os.environ['GXD_PDK_CONFIG'], files_path )
gnSchedulesParser.run( xml_file )

