#  -----------------------------------------------------------------------------
#  Gracenote (A Nielsen Company) Proprietary Information
#  This software is supplied under the terms of a license agreement or
#  nondisclosure agreement with Nielsen/Gracenote, Inc. and may not be copied
#  or disclosed except in accordance with the terms of that agreement.
#  Copyright(c) 2020-present. Gracenote, A Nielsen Company, All Rights Reserved.
#  -----------------------------------------------------------------------------

#  =============================================================================
#  Parses a given ON SourcePrograms XML file
#  Validates each record against fields configuration file
#  Writes validated records into TSV files for the endpoint
#  =============================================================================

import os
import sys

sys.path.append(os.path.join(sys.path[0],'..'))
import gxd_functions as gxd

xml_file 	= sys.argv[1]
files_path 	= sys.argv[2]

class GnSourceProgramsParser(gxd.GnEndpointParser):

	def parse(self, xml_file):

		use_natural_keys = (self.parser_options.get('natural_keys') != '0') 	# use natural keys by default

		# Override if specific option given
		if self.parser_options.get('natural_srcPrgSeasonId'):
			use_natural_keys = (self.parser_options.get('natural_srcPrgSeasonId') != '0') 	# use natural keys by default

		for p in self.xml_nodes(xml_file, "on/programs/program"):

			program = {}

			gnidsId 	= self.check( p, 'id' )
			updateId 	= self.check( p, 'updateId' )
			updateDate	= self.check( p, 'updateDate' )

			program['gnidsId'] 	 	= gnidsId
			program['updateId'] 	= updateId
			program['updateDate']	= updateDate

			program['TMSId']		= self.check( p, 'TMSId' )

			is_program_deleted 		= self.format_boolean( self.check( p, 'deleted' ))
			program['deleted'] 		= is_program_deleted

			program['desc_lang'] 	= self.check( p.find('descriptions'), 'lang' )

			program['progType'] 	= self.check( p.find('progType') )
			program['subType'] 		= self.check( p.find('subType') )

			program['origNetwork'] 	= self.check( p.find('originalNetwork') )
			program['partNumber']  	= self.check( p.find('partNumber') )

			# source program fields not existing in programs
			program['sourceDataProviderId'] 	= self.check( p.find('sourceDataProvider/id') )
			program['sourceDataProviderName'] 	= self.check( p.find('sourceDataProvider/name') )
			program['targetAudience'] 		= self.check( p.find('targetAudience') )
			program['productionStatus'] 	= self.check( p.find('productionStatus') )
			program['yearOfRelease'] 		= self.check( p.find('yearOfRelease') )

			# release date is currently part of releases/release xpath
			for pr in p.findall('releases/release'):
				program['releaseDate']	= self.check( pr.find('date') )

			# version specific fields
			program['version_releaseDate'] 		= self.check( p.find('versionInfo/releaseDate') )
			program['version_yearOfRelease'] 	= self.check( p.find('versionInfo/yearOfRelease') )
			program['version_colorCode'] 		= self.check( p.find('versionInfo/colorCode') )
			program['version_duration'] 		= self.check( p.find('versionInfo/duration') )


			if self.do_output_tsv:
				if not self.validate_and_write_tsv(program, 'sourceprogram', 'gnidsId'):
					p.clear()
					continue

			season_inc = 0
			for s in p.findall('seasons/season'):
				season = {}

				seasonGnId 		= self.check( s, 'seasonId' )

				if use_natural_keys:
					seasonId 	= seasonGnId
				else:
					season_inc 	+= 1
					seasonId 	= "{}-{}".format( gnidsId, str(season_inc) )		# season object composite primary key

				season['spSeasonId'] 	= seasonId
				season['updateId'] 		= updateId
				season['updateDate']	= updateDate
				season['gnidsId'] 	 	= gnidsId
				season['seasonId'] 		= seasonGnId

				#season['deleted'] 		= is_program_deleted
				#season['inactive'] 	= is_program_inactive

				season['nonSeason'] 		= self.format_boolean( self.check( s, 'nonSeason' ))
				season['season'] 			= self.check( s, 'seasonNumber' )
				season['ord'] 				= self.check( s, 'seasonOrdinal' )
				season['releaseDate']  		= self.check( s.find('releaseDate') )
				season['seasonFinale']  	= self.check( s.find('seasonFinale') )
				season['originalNetwork'] 	= self.check( s.find('originalNetwork') )

				if False: #self.do_parse_all:
					season['seasonYear'] 		= self.check( s, 'seasonYear' )
					season['origNetwork_prgSvcId']	= self.check( s.find('originalNetwork'), 'prgSvcId' )
					season['totalSeasonEpisodes']	= self.check( s.find('totalSeasonEpisodes') )
					season['showcard'] 	  		= self.format_boolean( self.check( s.find('showcard') ))
					season['officialURL']  		= self.check( s.find('officialURL') )
					season['title_lang'] 		= self.check( s.find('titles'), 'lang' )
					season['desc_lang']  		= self.check( s.find('descriptions'), 'lang' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(season, 'sourceprogram_season', 'gnidsId'):
						s.clear()
						continue


				for st in s.findall('titles/title'):

					title = {}

					title['updateId'] 		= updateId
					title['spSeasonId'] 	= seasonId
					title['gnidsId'] 		= gnidsId

					title['title']		= self.check( st )
					#title['size'] 		= self.check( st, 'size' )
					#title['type'] 		= self.check( st, 'type' )
					title['lang'] 		= self.check( st, 'lang' )
					#title['subType']	= self.check( st, 'subType' )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(title, 'sourceprogram_season_title', 'spSeasonId'):
							st.clear()
							continue

					if self.do_output_jsonl:
						[ title.pop(key) for key in ['spSeasonId','gnidsId','updateId'] ]
						season['titles'] = ( season.get('titles') or [] ) + [ title.copy() ]

					st.clear()

				for sd in s.findall('descriptions/desc'):

					description = {}

					description['updateId'] 	= updateId
					description['spSeasonId'] 	= seasonId
					description['gnidsId'] 		= gnidsId

					description['description'] 	= self.check( sd )
					description['size'] 		= self.check( sd, 'size' )
					#description['type'] 		= self.check( sd, 'type' )
					description['lang'] 		= self.check( sd, 'lang' )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(description, 'sourceprogram_season_desc', 'spSeasonId'):
							sd.clear()
							continue

					if self.do_output_jsonl:
						[ description.pop(key) for key in ['spSeasonId','gnidsId','updateId'] ]
						season['descriptions'] = ( season.get('descriptions') or [] ) + [ description.copy() ]

					sd.clear()


				for sc in s.findall('cast/member'):

					cast = {}

					cast['updateId'] 	= updateId
					cast['spSeasonId'] 	= seasonId
					cast['gnidsId'] 	= gnidsId

					cast['ord'] 			= self.check( sc, 'ord' )
					#cast['personId'] 		= self.check( sc, 'personId' )
					cast['characterName'] 	= self.check( sc.find('characterName') )
					cast['role'] 			= self.check( sc.find('role') )
					#cast['roleId'] 			= self.check( sc.find('role'), 'roleId' )
					#cast['nameId'] 			= self.check( sc.find('name'), 'nameId' )
					#cast['departed'] 	  	= self.format_boolean( self.check( sc.find('departed') ))
					cast['prefix'] 			= self.check( sc.find('name/prefix') )
					cast['first'] 			= self.check( sc.find('name/first') )
					cast['middle'] 			= self.check( sc.find('name/middle') )
					cast['last'] 			= self.check( sc.find('name/last') )
					cast['suffix'] 			= self.check( sc.find('name/suffix') )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(cast, 'sourceprogram_season_cast', 'spSeasonId'):
							sc.clear()
							continue

					if self.do_output_jsonl:
						[ cast.pop(key) for key in ['seasonId','gnidsId','updateId'] ]
						season['cast'] = ( season.get('cast') or [] ) + [ cast.copy() ]

					sc.clear()


				for sc in s.findall('crew/member'):

					crew = {}

					crew['updateId'] 	= updateId
					crew['spSeasonId'] 	= seasonId
					crew['gnidsId'] 	= gnidsId

					crew['ord'] 			= self.check( sc, 'ord' )
					#crew['personId'] 		= self.check( sc, 'personId' )
					crew['role'] 			= self.check( sc.find('role') )
					#crew['roleId'] 			= self.check( sc.find('role'), 'roleId' )
					#crew['nameId'] 			= self.check( sc.find('name'), 'nameId' )

					crew['prefix'] 			= self.check( sc.find('name/prefix') )
					crew['first'] 			= self.check( sc.find('name/first') )
					crew['middle'] 			= self.check( sc.find('name/middle') )
					crew['last'] 			= self.check( sc.find('name/last') )
					crew['suffix'] 			= self.check( sc.find('name/suffix') )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(crew, 'sourceprogram_season_crew', 'spSeasonId'):
							sc.clear()
							continue

					if self.do_output_jsonl:
						[ crew.pop(key) for key in ['seasonId','gnidsId','updateId'] ]
						season['crew'] = ( season.get('crew') or [] ) + [ crew.copy() ]

					sc.clear()


				for sx in s.findall('externalIds/externalId'):

					xid = {}

					xid['updateId'] 	= updateId
					xid['spSeasonId'] 	= seasonId
					xid['gnidsId'] 		= gnidsId

					xid['id']			= self.check( sx )
					xid['primary'] 		= self.check( sx, 'primary' )
					xid['provider'] 	= self.check( sx, 'provider' )
					xid['label'] 		= self.check( sx, 'label' )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(xid, 'sourceprogram_season_xid', 'spSeasonId'):
							sx.clear()
							continue

					if self.do_output_jsonl:
						[ xid.pop(key) for key in ['spSeasonId','gnidsId','updateId'] ]
						season['xids'] = ( season.get('xids') or [] ) + [ xid.copy() ]

					sx.clear()


				asset_inc = 0
				for sa in s.findall('assets/asset'):

					asset = {}

					assetId = self.check( sa, 'assetId' )
					asset_inc += 1
					seasonAssetId = "{0}-{1}".format( seasonId, asset_inc )

					asset['spAssetId'] 		= seasonAssetId
					asset['spSeasonId'] 	= seasonId
					asset['gnidsId'] 		= gnidsId
					asset['updateId'] 		= updateId
					asset['assetId'] 		= assetId

					asset['type'] 			= self.check( sa, 'type' )
					asset['expiredDate'] 	= self.check( sa, 'expiredDate' )
					asset['lastModified'] 	= self.check( sa, 'lastModified' )
					asset['created'] 		= self.check( sa, 'created' )
					asset['width'] 			= self.check( sa, 'width' )
					asset['height'] 		= self.check( sa, 'height' )
					asset['category'] 		= self.check( sa, 'category' )
					asset['ratio'] 			= self.check( sa, 'ratio' )
					asset['tier'] 			= self.check( sa, 'tier' )
					asset['seasonId'] 		= self.check( sa, 'seasonId' )
					asset['provider'] 		= self.check( sa, 'provider' )
					asset['primary'] 		= self.format_boolean( self.check( sa, 'primary' ))
					asset['deleted'] 		= self.format_boolean( self.check( sa, 'deleted' ))			
					asset['isConforming'] 	= self.format_boolean( self.check( sa, 'isConforming' ))
					asset['isBranded'] 		= self.format_boolean( self.check( sa, 'isBranded' ))

					asset['isSensitive'] 	= self.format_boolean( self.check( sa.find('isSensitive') ))
					asset['title'] 			= self.check( sa.find('title') )
					asset['URL'] 			= self.check( sa.find('URL') )
					asset['description'] 	= self.check( sa.find('descriptions/description') )
					asset['brandedDescription'] = self.check( sa.find('brandedDescription') )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(asset, 'sourceprogram_season_asset', 'spAssetId'):
							sa.clear()
							continue


					for sa_lang in sa.findall('languages/language'):

						lang = {}

						lang['spAssetId'] 	= seasonAssetId
						lang['gnidsId'] 	= gnidsId
						lang['updateId'] 	= updateId

						lang['lang'] 		= self.check( sa_lang )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(lang, 'sourceprogram_season_asset_language', 'spAssetId'):
								sa_lang.clear()
								continue

						if self.do_output_jsonl:
							#[ lang.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
							asset['languages'] = ( asset.get('languages') or [] ) + [ lang['lang'] ]

						sa_lang.clear()


					for sa_market in sa.findall('markets/market'):

						market = {}

						market['spAssetId'] 	= seasonAssetId
						market['gnidsId'] 		= gnidsId
						market['updateId'] 		= updateId
						
						market['market'] 		= self.check( sa_market )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(market, 'sourceprogram_season_asset_market', 'spAssetId'):
								sa_market.clear()
								continue

						if self.do_output_jsonl:
							#[ market.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
							asset['markets'] = ( asset.get('markets') or [] ) + [ market['market'] ]

						sa_market.clear()


					for sa_extattr in sa.findall('extendedAttributes/extendedAttribute'):

						ext_attr = {}

						ext_attr['spAssetId'] 	= seasonAssetId
						ext_attr['gnidsId'] 	= gnidsId
						ext_attr['updateId'] 	= updateId
						
						ext_attr['label'] 		= self.check( sa_extattr.find('label') )
						ext_attr['value'] 		= self.check( sa_extattr.find('value') )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(ext_attr, 'sourceprogram_season_asset_extattr', 'spAssetId'):
								sa_extattr.clear()
								continue

						if self.do_output_jsonl:
							[ ext_attr.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
							asset['ext_attributes'] = ( asset.get('ext_attributes') or [] ) + [ ext_attr.copy() ]

						sa_extattr.clear()


					for sa_tag in sa.findall('tags/tag'):

						tag = {}

						tag['spAssetId'] 	= seasonAssetId
						tag['gnidsId'] 		= gnidsId
						tag['updateId'] 	= updateId
						
						tag['tag'] 			= self.check( sa_tag )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(tag, 'sourceprogram_season_asset_tag', 'spAssetId'):
								sa_tag.clear()
								continue

						if self.do_output_jsonl:
							#[ tag.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
							asset['tags'] = ( asset.get('tags') or [] ) + [ tag['tag'] ]

						sa_tag.clear()


					for sa_identifier in sa.findall('identifiers/identifier'):

						# The XML schema here somewhat deviates from typical convention (items/item):
						# The actual elements (<id> tag) are nested under <identifier> tag which carries the type attribute

						sa_identifier_type = self.check( sa_identifier, 'type' )

						for sa_id in sa_identifier.findall('id'):

							asset_id = {}

							asset_id['spAssetId'] 		= seasonAssetId
							asset_id['updateId'] 		= updateId
							asset_id['gnidsId'] 		= gnidsId
							asset_id['type'] 			= sa_identifier_type
							asset_id['id'] 				= self.check( sa_id )

							if self.do_output_tsv:
								if not self.validate_and_write_tsv(asset_id, 'sourceprogram_season_asset_id', 'spAssetId'):
									sa_id.clear()
									continue

							if self.do_output_jsonl:
								[ asset_id.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
								asset['identifiers'] = ( asset.get('identifiers') or [] ) + [ asset_id.copy() ]

							sa_id.clear()

						sa_identifier.clear()


					if self.do_output_jsonl:
						[ asset.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
						season['assets'] = ( season.get('assets') or [] ) + [ asset.copy() ]

					sa.clear()

				if self.do_output_jsonl:
					self.write_jsonl(season, 'season')

				s.clear()	


			# Back to program child tables

			for pl in p.findall('versionInfo/presentationLabels/presentationLabel'):

				label = {}

				label['updateId'] 	= updateId
				label['gnidsId'] 	= gnidsId

				label['label'] 		= self.check( pl )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(label, 'sourceprogram_label', 'gnidsId'):
						pl.clear()
						continue

				if self.do_output_jsonl:
					[ label.pop(key) for key in ['gnidsId','updateId'] ]
					program['labels'] = ( program.get('labels') or [] ) + [ label.copy() ]

				pl.clear()


			for pvl in p.findall('versionInfo/labels/label'):

				ver_label = {}

				ver_label['updateId'] 	= updateId
				ver_label['gnidsId'] 	= gnidsId

				ver_label['label'] 		= self.check( pvl )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(ver_label, 'sourceprogram_versionlabel', 'gnidsId'):
						pvl.clear()
						continue

				if self.do_output_jsonl:
					[ ver_label.pop(key) for key in ['gnidsId','updateId'] ]
					program['version_labels'] = ( program.get('version_labels') or [] ) + [ ver_label.copy() ]

				pvl.clear()

			cat_inc = 0
			for pcat in p.findall('sourceDataProvider/catalogs/catalog'):

				catalog = {}
				cat_inc += 1

				catalogId 	= "{}-{}".format( gnidsId, str(cat_inc) )

				catalog['updateId'] 	= updateId
				catalog['gnidsId'] 		= gnidsId
				catalog['spCatalogId'] 	= catalogId

				catalog['catalogId'] 	= self.check( pcat, 'id' )
				catalog['catalogName'] 	= self.check( pcat, 'name' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(catalog, 'sourceprogram_catalog', 'gnidsId'):
						pcat.clear()
						continue

				for pcl in pcat.findall('languages/language'):

					lang = {}

					lang['updateId'] 		= updateId
					lang['gnidsId'] 		= gnidsId
					lang['spCatalogId'] 	= catalogId

					lang['lang'] 			= self.check( pcl )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(lang, 'sourceprogram_catalog_language', 'spCatalogId'):
							pcl.clear()
							continue
					
					if self.do_output_jsonl:
						[ lang.pop(key) for key in ['gnidsId','updateId','spCatalogId'] ]
						catalog['languages'] = ( catalog.get('languages') or [] ) + [ lang.copy() ]

					pcl.clear()

				for pcm in pcat.findall('markets/market'):

					market = {}

					market['updateId'] 		= updateId
					market['gnidsId'] 		= gnidsId
					market['spCatalogId'] 	= catalogId

					market['market'] 		= self.check( pcm )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(market, 'sourceprogram_catalog_market', 'spCatalogId'):
							pcm.clear()
							continue
					
					if self.do_output_jsonl:
						[ market.pop(key) for key in ['gnidsId','updateId','spCatalogId'] ]
						catalog['markets'] = ( catalog.get('markets') or [] ) + [ market.copy() ]

					pcm.clear()


				if self.do_output_jsonl:
					[ catalog.pop(key) for key in ['gnidsId','updateId','spCatalogId'] ]
					program['catalogs'] = ( program.get('catalogs') or [] ) + [ catalog.copy() ]

				pcat.clear()


			for pt in p.findall('titles/title'):

				title = {}

				title['updateId'] 	= updateId
				title['gnidsId'] 	= gnidsId

				title['title'] 		= self.check( pt )
				title['size'] 		= self.check( pt, 'size' )
				title['type'] 		= self.check( pt, 'type' )
				title['lang'] 		= self.check( pt, 'lang' )
				title['subType'] 	= self.check( pt, 'subType' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(title, 'sourceprogram_title', 'gnidsId'):
						pt.clear()
						continue

				if self.do_output_jsonl:
					[ title.pop(key) for key in ['gnidsId','updateId'] ]
					program['titles'] = ( program.get('titles') or [] ) + [ title.copy() ]

				pt.clear()


			for pd in p.findall('descriptions/desc'):

				description = {}

				description['updateId'] 	= updateId
				description['gnidsId'] 		= gnidsId

				description['description']	= self.check( pd )
				description['size']			= self.check( pd, 'size' )
				description['type']			= self.check( pd, 'type' )
				description['lang']			= self.check( pd, 'lang' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(description, 'sourceprogram_description', 'gnidsId'):
						pd.clear()
						continue

				if self.do_output_jsonl:
					[ description.pop(key) for key in ['gnidsId','updateId'] ]
					program['descriptions'] = ( program.get('descriptions') or [] ) + [ description.copy() ]

				pd.clear()


			for pc in p.findall('cast/member'):

				cast = {}

				cast['updateId'] 	= updateId
				cast['gnidsId'] 		= gnidsId

				cast['ord'] 			= self.check( pc, 'ord' )
				#cast['personId'] 		= self.check( pc, 'personId' )
				cast['characterName'] 	= self.check( pc.find('characterName') )
				cast['role'] 			= self.check( pc.find('role') )
				#cast['roleId'] 			= self.check( pc.find('role'), 'roleId' )
				#cast['nameId'] 			= self.check( pc.find('name'), 'nameId' )
				#cast['departed'] 	  	= self.format_boolean( self.check( pc.find('departed') ))
				cast['prefix'] 			= self.check( pc.find('name/prefix') )
				cast['first'] 			= self.check( pc.find('name/first') )
				cast['middle'] 			= self.check( pc.find('name/middle') )
				cast['last'] 			= self.check( pc.find('name/last') )
				cast['suffix'] 			= self.check( pc.find('name/suffix') )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(cast, 'sourceprogram_cast', 'gnidsId'):
						pc.clear()
						continue

				if self.do_output_jsonl:
					[ cast.pop(key) for key in ['gnidsId','updateId'] ]
					program['cast'] = ( program.get('cast') or [] ) + [ cast.copy() ]

				pc.clear()


			for pc in p.findall('crew/member'):

				crew = {}

				crew['updateId'] 	= updateId
				crew['gnidsId'] 		= gnidsId

				crew['ord'] 			= self.check( pc, 'ord' )
				#crew['personId'] 		= self.check( pc, 'personId' )
				crew['role'] 			= self.check( pc.find('role') )
				#crew['roleId'] 			= self.check( pc.find('role'), 'roleId' )
				#crew['nameId'] 			= self.check( pc.find('name'), 'nameId' )
				crew['prefix'] 			= self.check( pc.find('name/prefix') )
				crew['first'] 			= self.check( pc.find('name/first') )
				crew['middle'] 			= self.check( pc.find('name/middle') )
				crew['last'] 			= self.check( pc.find('name/last') )
				crew['suffix'] 			= self.check( pc.find('name/suffix') )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(crew, 'sourceprogram_crew', 'gnidsId'):
						pc.clear()
						continue

				if self.do_output_jsonl:
					[ crew.pop(key) for key in ['gnidsId','updateId'] ]
					program['crew'] = ( program.get('crew') or [] ) + [ crew.copy() ]

				pc.clear()

			for pc in p.findall('countries/country'):

				country = {}

				country['updateId'] 	= updateId
				country['gnidsId'] 		= gnidsId

				country['country'] 		= self.check( pc )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(country, 'sourceprogram_country', 'gnidsId'):
						pc.clear()
						continue

				if self.do_output_jsonl:
					#[ country.pop(key) for key in ['gnidsId','updateId'] ]
					program['countries'] = ( program.get('countries') or [] ) + [ country['country'] ]

				pc.clear()


			genre_inc = 0
			for pg in p.findall('genres/genre'):

				genre = {}
				genre_inc += 1

				genre['updateId'] 	= updateId
				genre['gnidsId'] 	= gnidsId

				genre['genreOrd']	= genre_inc
				genre['genre']		= self.check( pg )
				#genre['genreId'] 	= self.check( pg, 'genreId' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(genre, 'sourceprogram_genre', 'gnidsId'):
						pg.clear()
						continue

				if self.do_output_jsonl:
					[ genre.pop(key) for key in ['gnidsId','updateId'] ]
					program['genres'] = ( program.get('genres') or [] ) + [ genre.copy() ]

				pg.clear()


			rating_id = 0
			for pr in p.findall('ratings/rating'):

				rating = {}

				rating_id += 1
				ratingId = "{0}-{1}".format( gnidsId, str(rating_id))		

				rating['spRatingId'] 	= ratingId
				rating['updateId'] 		= updateId
				rating['gnidsId'] 		= gnidsId

				rating['exempt'] 		= self.format_boolean( self.check( pr, 'exempt' ))
				rating['rating_code'] 	= self.check( pr, 'code' )
				#rating['ratingBodyId'] 	= self.check( pr, 'ratingBodyId' )
				rating['ratingsBody'] 	= self.check( pr, 'ratingsBody' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(rating, 'sourceprogram_rating', 'gnidsId'):
						pr.clear()
						continue


				for pra in pr.findall('advisories/advisory'):

					advisory = {}

					advisory['spRatingId']		= ratingId
					advisory['updateId'] 		= updateId
					advisory['gnidsId'] 		= gnidsId

					advisory['advisory']		= self.check( pra )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(advisory, 'sourceprogram_rating_advisory', 'gnidsId'):
							pra.clear()
							continue

					if self.do_output_jsonl:
						[ advisory.pop(key) for key in ['gnidsId','updateId'] ]
						rating['advisories'] = ( rating.get('advisories') or [] ) + [ advisory.copy() ]

					pra.clear()


				if self.do_output_jsonl:
					[ rating.pop(key) for key in ['gnidsId','updateId'] ]
					program['ratings'] = ( program.get('ratings') or [] ) + [ rating.copy() ]

				pr.clear()



			if False:		# not part of Source Programs schema
				for pqr in p.findall('ratings/qualityRating'):

					qualityrating = {}

					qualityrating['updateId'] 	= updateId
					qualityrating['gnidsId'] 	= gnidsId

					qualityrating['ratingsBody'] 	= self.check( pqr, 'ratingsBody' )
					qualityrating['numVotes'] 		= self.check( pqr, 'numVotes' )
					qualityrating['value'] 			= self.check( pqr, 'value' )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(qualityrating, 'sourceprogram_qualityrating', 'gnidsId'):
							pqr.clear()
							continue

					if self.do_output_jsonl:
						[ qualityrating.pop(key) for key in ['gnidsId','updateId'] ]
						program['quality_ratings'] = ( program.get('quality_ratings') or [] ) + [ qualityrating.copy() ]

					pqr.clear()



			pmi = p.find('movieInfo')
			if pmi is not None:

				movieinfo = [    ('yearOfRelease', 		'yearOfRelease')
								,('officialURL', 		'officialURL')
								,('productionCompany', 	'productionCompanies/name')
								,('soundMix', 			'soundMixes/soundMix')
								,('pictureFormat', 		'pictureFormats/pictureFormat')		]

				for key, xpath in movieinfo:
					for pmi_value in pmi.findall(xpath):

						mvinfo = {}

						mvinfo['updateId'] 	= updateId
						mvinfo['gnidsId'] 	= gnidsId

						mvinfo['key']		= key
						mvinfo['value'] 	= self.check( pmi_value )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(mvinfo, 'sourceprogram_mvinfo', 'gnidsId'):
								pmi_value.clear()
								continue

						if self.do_output_jsonl:
							[ mvinfo.pop(key) for key in ['gnidsId','updateId'] ]
							program['movieInfo'] = ( program.get('movieInfo') or [] ) + [ mvinfo.copy() ]

						pmi_value.clear()


				if False:		# not part of Source Programs schema
					for pmit in p.findall('movieInfo/trailers/trailer'):

						movietrailer = {}

						movietrailer['updateId'] 	= updateId
						movietrailer['gnidsId'] 		= gnidsId

						movietrailer['format']		= self.check( pmit.find('format') )
						movietrailer['URL'] 		= self.check( pmit.find('URL') )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(movietrailer, 'sourceprogram_mvtrailer', 'gnidsId'):
								pmit.clear()
								continue

						if self.do_output_jsonl:
							[ movietrailer.pop(key) for key in ['gnidsId','updateId'] ]
							program['movieTrailers'] = ( program.get('movieTrailers') or [] ) + [ movietrailer.copy() ]

						pmit.clear()

				pmi.clear()


			episodeinfo_id = 0
			for pei in p.findall('episodeInfo'):

				episodeinfo = {}

				episodeinfo_id += 1
				episodeInfoId = "{0}-{1}".format( gnidsId, str(episodeinfo_id) )

				episodeinfo['spEpisodeInfoId'] 	= episodeInfoId
				episodeinfo['gnidsId'] 			= gnidsId
				episodeinfo['updateId'] 		= updateId

				title 							= pei.find('title')
				episodeinfo['title']			= self.check( title )
				episodeinfo['size'] 			= self.check( title, 'size' )
				episodeinfo['type'] 			= self.check( title, 'type' )
				episodeinfo['lang'] 			= self.check( title, 'lang' )

				episodeinfo['season'] 			= self.check( pei, 'season' )
				episodeinfo['number'] 			= self.check( pei, 'number' )
				episodeinfo['ord'] 				= self.check( pei, 'ordinal' )
				episodeinfo['synNum']			= self.check( pei.find('synNum') )
				episodeinfo['industryNumber'] 	= self.check( pei.find('industryNumber') )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(episodeinfo, 'sourceprogram_epinfo', 'gnidsId'):
						pei.clear()
						if title is not None:
							title.clear()
						continue

				if self.do_output_jsonl:
					[ episodeinfo.pop(key) for key in ['spEpisodeInfoId','gnidsId','updateId'] ]
					program['episodeInfo'] = ( program.get('episodeInfo') or [] ) + [ episodeinfo.copy() ]

				pei.clear()
				if title is not None:
					title.clear()


			for px in p.findall('externalIds/externalId'):

				xid = {}

				xid['updateId'] 	= updateId
				xid['gnidsId'] 		= gnidsId

				xid['id'] 		= self.check( px )
				xid['primary'] 	= self.check( px, 'primary' )
				xid['provider'] = self.check( px, 'provider' )
				xid['label'] 	= self.check( px, 'label' )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(xid, 'sourceprogram_xid', 'gnidsId'):
						px.clear()
						continue

				px.clear()


			asset_inc = 0
			for pa in p.findall('assets/asset'):

				asset = {}

				assetId = self.check( pa, 'assetId' )
				asset_inc += 1
				programAssetId = "{0}-{1}".format( gnidsId, asset_inc )

				asset['spAssetId'] 		= programAssetId
				asset['gnidsId'] 		= gnidsId
				asset['updateId'] 		= updateId
				asset['assetId'] 		= assetId

				asset['type'] 			= self.check( pa, 'type' )
				asset['expiredDate'] 	= self.check( pa, 'expiredDate' )
				asset['lastModified'] 	= self.check( pa, 'lastModified' )
				asset['created'] 		= self.check( pa, 'created' )
				asset['width'] 			= self.check( pa, 'width' )
				asset['height'] 		= self.check( pa, 'height' )
				asset['category'] 		= self.check( pa, 'category' )
				asset['ratio'] 			= self.check( pa, 'ratio' )
				asset['tier'] 			= self.check( pa, 'tier' )
				asset['seasonId'] 		= self.check( pa, 'seasonId' )
				asset['provider'] 		= self.check( pa, 'provider' )
				asset['primary'] 		= self.format_boolean( self.check( pa, 'primary' ))
				asset['deleted'] 		= self.format_boolean( self.check( pa, 'deleted' ))			
				asset['isConforming'] 	= self.format_boolean( self.check( pa, 'isConforming' ))
				asset['isBranded'] 		= self.format_boolean( self.check( pa, 'isBranded' ))

				asset['isSensitive'] 	= self.format_boolean( self.check( pa.find('isSensitive') ))
				asset['title'] 			= self.check( pa.find('title') )
				asset['URL'] 			= self.check( pa.find('URL') )
				asset['description'] 	= self.check( pa.find('descriptions/description') )
				asset['brandedDescription'] = self.check( pa.find('brandedDescription') )

				if self.do_output_tsv:
					if not self.validate_and_write_tsv(asset, 'sourceprogram_asset', 'spAssetId'):
						pa.clear()
						continue


				for pa_lang in pa.findall('languages/language'):

					lang = {}

					lang['spAssetId'] 	= programAssetId
					lang['gnidsId'] 	= gnidsId
					lang['updateId'] 	= updateId

					lang['lang'] 		= self.check( pa_lang )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(lang, 'sourceprogram_asset_language', 'spAssetId'):
							pa_lang.clear()
							continue

					if self.do_output_jsonl:
						#[ lang.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
						asset['languages'] = ( asset.get('languages') or [] ) + [ lang['lang'] ]

					pa_lang.clear()


				for pa_market in pa.findall('markets/market'):

					market = {}

					market['spAssetId'] 	= programAssetId
					market['gnidsId'] 		= gnidsId
					market['updateId'] 		= updateId
					
					market['market'] 		= self.check( pa_market )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(market, 'sourceprogram_asset_market', 'spAssetId'):
							pa_market.clear()
							continue

					if self.do_output_jsonl:
						#[ market.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
						asset['markets'] = ( asset.get('markets') or [] ) + [ market['market'] ]

					pa_market.clear()


				for pa_extattr in pa.findall('extendedAttributes/extendedAttribute'):

					ext_attr = {}

					ext_attr['spAssetId'] 	= programAssetId
					ext_attr['gnidsId'] 	= gnidsId
					ext_attr['updateId'] 	= updateId
					
					ext_attr['label'] 		= self.check( pa_extattr.find('label') )
					ext_attr['value'] 		= self.check( pa_extattr.find('value') )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(ext_attr, 'sourceprogram_asset_extattr', 'spAssetId'):
							pa_extattr.clear()
							continue

					if self.do_output_jsonl:
						[ ext_attr.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
						asset['ext_attributes'] = ( asset.get('ext_attributes') or [] ) + [ ext_attr.copy() ]

					pa_extattr.clear()


				for pa_tag in pa.findall('tags/tag'):

					tag = {}

					tag['spAssetId'] 	= programAssetId
					tag['gnidsId'] 		= gnidsId
					tag['updateId'] 	= updateId
					
					tag['market'] 		= self.check( pa_tag )

					if self.do_output_tsv:
						if not self.validate_and_write_tsv(tag, 'sourceprogram_asset_market', 'spAssetId'):
							pa_tag.clear()
							continue

					if self.do_output_jsonl:
						#[ market.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
						asset['tags'] = ( asset.get('tags') or [] ) + [ tag['tag'] ]

					pa_tag.clear()


				for pa_identifier in pa.findall('identifiers/identifier'):

					# The XML schema here somewhat deviates from typical convention (items/item):
					# The actual elements (<id> tag) are nested under <identifier> tag which carries the type attribute

					pa_identifier_type = self.check( pa_identifier, 'type' )

					for pa_id in pa_identifier.findall('id'):

						asset_id = {}

						asset_id['spAssetId'] 		= programAssetId
						asset_id['gnidsId'] 		= gnidsId
						asset_id['updateId'] 		= updateId
						asset_id['type'] 			= pa_identifier_type
						asset_id['id'] 				= self.check( pa_id )

						if self.do_output_tsv:
							if not self.validate_and_write_tsv(asset_id, 'sourceprogram_asset_id', 'spAssetId'):
								pa_id.clear()
								continue

						if self.do_output_jsonl:
							[ asset_id.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
							asset['identifiers'] = ( asset.get('identifiers') or [] ) + [ asset_id.copy() ]

						pa_id.clear()

					pa_identifier.clear()

				if self.do_output_jsonl:
					[ asset.pop(key) for key in ['spAssetId','gnidsId','updateId'] ]
					program['assets'] = ( program.get('assets') or [] ) + [ asset.copy() ]

				pa.clear()




			if self.do_output_jsonl:
				self.write_jsonl(program, 'program')

			p.clear()


# ================== ENDPOINT SPECIFIC INIT CODE BEGIN ==================
# Here insert the endpoint-specific code from schema/release/<product>/parse_.py


endpoint = 'SourcePrograms'

entities = [
	 'sourceprogram' 
]

tables = [
	 'sourceprogram' 
	,'sourceprogram_label' 
	,'sourceprogram_versionlabel' 
	,'sourceprogram_catalog' 
	,'sourceprogram_catalog_language' 
	,'sourceprogram_catalog_market' 
	,'sourceprogram_title' 
	,'sourceprogram_description' 
	,'sourceprogram_cast' 
	,'sourceprogram_crew' 
	,'sourceprogram_country' 
	,'sourceprogram_genre' 
	,'sourceprogram_rating' 
	,'sourceprogram_rating_advisory' 
	,'sourceprogram_mvinfo' 
	,'sourceprogram_epinfo' 
	,'sourceprogram_season' 
	,'sourceprogram_season_title' 
	,'sourceprogram_season_desc' 
	,'sourceprogram_season_cast' 
	,'sourceprogram_season_crew' 
	,'sourceprogram_season_asset' 
	,'sourceprogram_season_asset_id' 
	,'sourceprogram_season_asset_language' 
	,'sourceprogram_season_asset_market' 
	,'sourceprogram_season_asset_extattr' 
	,'sourceprogram_season_asset_tag' 
	,'sourceprogram_season_xid' 
	,'sourceprogram_asset' 
	,'sourceprogram_asset_id' 
	,'sourceprogram_asset_language' 
	,'sourceprogram_asset_market' 
	,'sourceprogram_asset_extattr' 
	,'sourceprogram_asset_tag' 
	,'sourceprogram_xid' 

]

fields = {
	 'sourceprogram': ['gnidsId', 'updateId', 'updateDate', 'deleted', 'TMSId', 'desc_lang', 'progType', 'subType', 'sourceDataProviderId', 'sourceDataProviderName', 'origNetwork', 'targetAudience', 'productionStatus', 'releaseDate', 'yearOfRelease', 'partNumber', 'version_yearOfRelease', 'version_releaseDate', 'version_colorCode', 'version_duration', 'MD5', 'ingestId', 'batchUpdateOrd']
	,'sourceprogram_label': ['gnidsId', 'updateId', 'label', 'MD5', 'ingestId']
	,'sourceprogram_versionlabel': ['gnidsId', 'updateId', 'label', 'MD5', 'ingestId']
	,'sourceprogram_catalog': ['spCatalogId', 'gnidsId', 'updateId', 'catalogId', 'catalogName', 'MD5', 'ingestId']
	,'sourceprogram_catalog_language': ['spCatalogId', 'gnidsId', 'updateId', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_catalog_market': ['spCatalogId', 'gnidsId', 'updateId', 'market', 'MD5', 'ingestId']
	,'sourceprogram_title': ['gnidsId', 'updateId', 'title', 'size', 'type', 'subType', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_description': ['gnidsId', 'updateId', 'description', 'size', 'type', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_cast': ['gnidsId', 'updateId', 'ord', 'characterName', 'role', 'prefix', 'first', 'middle', 'last', 'suffix', 'MD5', 'ingestId']
	,'sourceprogram_crew': ['gnidsId', 'updateId', 'ord', 'role', 'prefix', 'first', 'middle', 'last', 'suffix', 'MD5', 'ingestId']
	,'sourceprogram_country': ['gnidsId', 'updateId', 'country', 'MD5', 'ingestId']
	,'sourceprogram_genre': ['gnidsId', 'updateId', 'genreOrd', 'genre', 'MD5', 'ingestId']
	,'sourceprogram_rating': ['spRatingId', 'gnidsId', 'updateId', 'exempt', 'rating_code', 'ratingsBody', 'MD5', 'ingestId']
	,'sourceprogram_rating_advisory': ['spRatingId', 'gnidsId', 'updateId', 'advisory', 'MD5', 'ingestId']
	,'sourceprogram_mvinfo': ['gnidsId', 'updateId', 'key', 'value', 'MD5', 'ingestId']
	,'sourceprogram_epinfo': ['spEpisodeInfoId', 'gnidsId', 'updateId', 'title', 'size', 'type', 'lang', 'season', 'number', 'ord', 'synNum', 'industryNumber', 'MD5', 'ingestId']
	,'sourceprogram_season': ['spSeasonId', 'gnidsId', 'updateId', 'nonSeason', 'seasonId', 'season', 'ord', 'releaseDate', 'seasonFinale', 'originalNetwork', 'MD5', 'ingestId']
	,'sourceprogram_season_title': ['spSeasonId', 'gnidsId', 'updateId', 'title', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_season_desc': ['spSeasonId', 'gnidsId', 'updateId', 'description', 'size', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_season_cast': ['spSeasonId', 'gnidsId', 'updateId', 'ord', 'characterName', 'role', 'prefix', 'first', 'middle', 'last', 'suffix', 'MD5', 'ingestId']
	,'sourceprogram_season_crew': ['spSeasonId', 'gnidsId', 'updateId', 'ord', 'role', 'prefix', 'first', 'middle', 'last', 'suffix', 'MD5', 'ingestId']
	,'sourceprogram_season_asset': ['spAssetId', 'spSeasonId', 'gnidsId', 'updateId', 'assetUpdateId', 'assetId', 'deleted', 'created', 'lastModified', 'expiredDate', 'type', 'width', 'height', 'primary', 'category', 'ratio', 'tier', 'seasonId', 'provider', 'isConforming', 'isBranded', 'title', 'URL', 'description', 'brandedDescription', 'isSensitive', 'MD5', 'ingestId']
	,'sourceprogram_season_asset_id': ['spAssetId', 'gnidsId', 'updateId', 'id', 'type', 'MD5', 'ingestId']
	,'sourceprogram_season_asset_language': ['spAssetId', 'gnidsId', 'updateId', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_season_asset_market': ['spAssetId', 'gnidsId', 'updateId', 'market', 'MD5', 'ingestId']
	,'sourceprogram_season_asset_extattr': ['spAssetId', 'gnidsId', 'updateId', 'label', 'value', 'MD5', 'ingestId']
	,'sourceprogram_season_asset_tag': ['spAssetId', 'gnidsId', 'updateId', 'tag', 'MD5', 'ingestId']
	,'sourceprogram_season_xid': ['spSeasonId', 'gnidsId', 'updateId', 'id', 'primary', 'provider', 'label', 'MD5', 'ingestId']
	,'sourceprogram_asset': ['spAssetId', 'gnidsId', 'updateId', 'assetUpdateId', 'assetId', 'deleted', 'created', 'lastModified', 'expiredDate', 'type', 'width', 'height', 'primary', 'category', 'ratio', 'tier', 'seasonId', 'provider', 'isConforming', 'isBranded', 'title', 'URL', 'description', 'brandedDescription', 'isSensitive', 'MD5', 'ingestId']
	,'sourceprogram_asset_id': ['spAssetId', 'gnidsId', 'updateId', 'id', 'type', 'MD5', 'ingestId']
	,'sourceprogram_asset_language': ['spAssetId', 'gnidsId', 'updateId', 'lang', 'MD5', 'ingestId']
	,'sourceprogram_asset_market': ['spAssetId', 'gnidsId', 'updateId', 'market', 'MD5', 'ingestId']
	,'sourceprogram_asset_extattr': ['spAssetId', 'gnidsId', 'updateId', 'label', 'value', 'MD5', 'ingestId']
	,'sourceprogram_asset_tag': ['spAssetId', 'gnidsId', 'updateId', 'tag', 'MD5', 'ingestId']
	,'sourceprogram_xid': ['gnidsId', 'updateId', 'id', 'primary', 'provider', 'label', 'MD5', 'ingestId']
}


# =================== ENDPOINT SPECIFIC INIT CODE END ===================


gnSourceProgramsParser = GnSourceProgramsParser( endpoint, entities, tables, fields )
gnSourceProgramsParser.set_config( os.environ['GXD_PDK_DIR'], os.environ['GXD_PDK_CONFIG'], files_path )
gnSourceProgramsParser.run( xml_file )
